from asyncio import Future, Queue, ensure_future, sleep
from inspect import isawaitable
from typing import Any, AsyncIterator, Callable, Optional, Set

try:
    from asyncio import get_running_loop
except ImportError:
    from asyncio import get_event_loop as get_running_loop  # Python < 3.7


__all__ = ["SimplePubSub", "SimplePubSubIterator"]


class SimplePubSub:
    """A very simple publish-subscript system.

    Creates an AsyncIterator from an EventEmitter.

    Useful for mocking a PubSub system for tests.
    """

    subscribers: Set[Callable]

    def __init__(self) -> None:
        self.subscribers = set()

    def emit(self, event: Any) -> bool:
        """Emit an event."""
        for subscriber in self.subscribers:
            result = subscriber(event)
            if isawaitable(result):
                ensure_future(result)
        return bool(self.subscribers)

    def get_subscriber(
        self, transform: Optional[Callable] = None
    ) -> "SimplePubSubIterator":
        return SimplePubSubIterator(self, transform)


class SimplePubSubIterator(AsyncIterator):
    def __init__(self, pubsub: SimplePubSub, transform: Optional[Callable]) -> None:
        self.pubsub = pubsub
        self.transform = transform
        self.pull_queue: Queue[Future] = Queue()
        self.push_queue: Queue[Any] = Queue()
        self.listening = True
        pubsub.subscribers.add(self.push_value)

    def __aiter__(self) -> "SimplePubSubIterator":
        return self

    async def __anext__(self) -> Any:
        if not self.listening:
            raise StopAsyncIteration
        await sleep(0)
        if not self.push_queue.empty():
            return await self.push_queue.get()
        future = get_running_loop().create_future()
        await self.pull_queue.put(future)
        return future

    async def aclose(self) -> None:
        if self.listening:
            await self.empty_queue()

    async def empty_queue(self) -> None:
        self.listening = False
        self.pubsub.subscribers.remove(self.push_value)
        while not self.pull_queue.empty():
            future = await self.pull_queue.get()
            future.cancel()
        while not self.push_queue.empty():
            await self.push_queue.get()

    async def push_value(self, event: Any) -> None:
        value = event if self.transform is None else self.transform(event)
        if self.pull_queue.empty():
            await self.push_queue.put(value)
        else:
            (await self.pull_queue.get()).set_result(value)



from typing import Any, Tuple, Union

__all__ = [
    "Description",
    "is_description",
    "register_description",
    "unregister_description",
]


class Description:
    """Type checker for human readable descriptions.

    By default, only ordinary strings are accepted as descriptions,
    but you can register() other classes that will also be allowed,
    e.g. to support lazy string objects that are evaluated only at runtime.
    If you register(object), any object will be allowed as description.
    """

    bases: Union[type, Tuple[type, ...]] = str

    @classmethod
    def isinstance(cls, obj: Any) -> bool:
        return isinstance(obj, cls.bases)

    @classmethod
    def register(cls, base: type) -> None:
        """Register a class that shall be accepted as a description."""
        if not isinstance(base, type):
            raise TypeError("Only types can be registered.")
        if base is object:
            cls.bases = object
        elif cls.bases is object:
            cls.bases = base
        elif not isinstance(cls.bases, tuple):
            if base is not cls.bases:
                cls.bases = (cls.bases, base)
        elif base not in cls.bases:
            cls.bases += (base,)

    @classmethod
    def unregister(cls, base: type) -> None:
        """Unregister a class that shall no more be accepted as a description."""
        if not isinstance(base, type):
            raise TypeError("Only types can be unregistered.")
        if isinstance(cls.bases, tuple):
            if base in cls.bases:
                cls.bases = tuple(b for b in cls.bases if b is not base)
            if not cls.bases:
                cls.bases = object
            elif len(cls.bases) == 1:
                cls.bases = cls.bases[0]
        elif cls.bases is base:
            cls.bases = object


is_description = Description.isinstance
register_description = Description.register
unregister_description = Description.unregister



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import ProvidedRequiredArgumentsRule
from graphql.validation.rules.provided_required_arguments import (
    ProvidedRequiredArgumentsOnDirectivesRule,
)

from .harness import assert_validation_errors, assert_sdl_validation_errors

assert_errors = partial(assert_validation_errors, ProvidedRequiredArgumentsRule)

assert_valid = partial(assert_errors, errors=[])

assert_sdl_errors = partial(
    assert_sdl_validation_errors, ProvidedRequiredArgumentsOnDirectivesRule
)

assert_sdl_valid = partial(assert_sdl_errors, errors=[])


def describe_validate_provided_required_arguments():
    def ignores_unknown_arguments():
        assert_valid(
            """
            {
              dog {
                isHouseTrained(unknownArgument: true)
              }
            }"""
        )

    def describe_valid_non_nullable_value():
        def arg_on_optional_arg():
            assert_valid(
                """
                {
                  dog {
                    isHouseTrained(atOtherHomes: true)
                  }
                }"""
            )

        def no_arg_on_optional_arg():
            assert_valid(
                """
                {
                  dog {
                    isHouseTrained
                  }
                }"""
            )

        def no_arg_on_non_null_field_with_default():
            assert_valid(
                """
                {
                  complicatedArgs {
                    nonNullFieldWithDefault
                  }
                }"""
            )

        def multiple_args():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleReqs(req1: 1, req2: 2)
                  }
                }
                """
            )

        def multiple_args_reverse_order():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleReqs(req2: 2, req1: 1)
                  }
                }
                """
            )

        def no_args_on_multiple_optional():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOpts
                  }
                }
                """
            )

        def one_arg_on_multiple_optional():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOpts(opt1: 1)
                  }
                }
                """
            )

        def second_arg_on_multiple_optional():
            assert_valid(
                """
                {
                    complicatedArgs {
                        multipleOpts(opt2: 1)
                    }
                }
                """
            )

        def multiple_required_args_on_mixed_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOptAndReq(req1: 3, req2: 4)
                  }
                }
                """
            )

        def multiple_required_and_one_optional_arg_on_mixed_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOptAndReq(req1: 3, req2: 4, opt1: 5)
                  }
                }
                """
            )

        def all_required_and_optional_args_on_mixed_list():
            assert_valid(
                """
                {
                  complicatedArgs {
                    multipleOptAndReq(req1: 3, req2: 4, opt1: 5, opt2: 6)
                  }
                }
                """
            )

    def describe_invalid_non_nullable_value():
        def missing_one_non_nullable_argument():
            assert_errors(
                """
                {
                  complicatedArgs {
                    multipleReqs(req2: 2)
                  }
                }
                """,
                [
                    {
                        "message": "Field 'multipleReqs' argument 'req1'"
                        " of type 'Int!' is required, but it was not provided.",
                        "locations": [(4, 21)],
                    },
                ],
            )

        def missing_multiple_non_nullable_arguments():
            assert_errors(
                """
                {
                  complicatedArgs {
                    multipleReqs
                  }
                }
                """,
                [
                    {
                        "message": "Field 'multipleReqs' argument 'req1'"
                        " of type 'Int!' is required, but it was not provided.",
                        "locations": [(4, 21)],
                    },
                    {
                        "message": "Field 'multipleReqs' argument 'req2'"
                        " of type 'Int!' is required, but it was not provided.",
                        "locations": [(4, 21)],
                    },
                ],
            )

        def incorrect_value_and_missing_argument():
            assert_errors(
                """
                {
                  complicatedArgs {
                    multipleReqs(req1: "one")
                  }
                }
                """,
                [
                    {
                        "message": "Field 'multipleReqs' argument 'req2'"
                        " of type 'Int!' is required, but it was not provided.",
                        "locations": [(4, 21)],
                    },
                ],
            )

    def describe_directive_arguments():
        def ignores_unknown_directives():
            assert_valid(
                """
                {
                  dog @unknown
                }
                """
            )

        def with_directives_of_valid_type():
            assert_valid(
                """
                {
                  dog @include(if: true) {
                    name
                  }
                  human @skip(if: false) {
                    name
                  }
                }
                """
            )

        def with_directive_with_missing_types():
            assert_errors(
                """
                {
                  dog @include {
                    name @skip
                  }
                }
                """,
                [
                    {
                        "message": "Directive '@include' argument 'if' of type"
                        " 'Boolean!' is required, but it was not provided.",
                        "locations": [(3, 23)],
                    },
                    {
                        "message": "Directive '@skip' argument 'if' of type"
                        " 'Boolean!' is required, but it was not provided.",
                        "locations": [(4, 26)],
                    },
                ],
            )

    def describe_within_sdl():
        def missing_optional_args_on_directive_defined_inside_sdl():
            assert_sdl_valid(
                """
                type Query {
                foo: String @test
                }

                directive @test(arg1: String, arg2: String! = "") on FIELD_DEFINITION
                """
            )

        def missing_arg_on_directive_defined_inside_sdl():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @test
                }

                directive @test(arg: String!) on FIELD_DEFINITION
                """,
                [
                    {
                        "message": "Directive '@test' argument 'arg' of type"
                        " 'String!' is required, but it was not provided.",
                        "locations": [(3, 31)],
                    },
                ],
            )

        def missing_arg_on_standard_directive():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @include
                }
                """,
                [
                    {
                        "message": "Directive '@include' argument 'if' of type"
                        " 'Boolean!' is required, but it was not provided.",
                        "locations": [(3, 31)],
                    },
                ],
            )

        def missing_arg_on_overridden_standard_directive():
            assert_sdl_errors(
                """
                type Query {
                  foo: String @deprecated
                }
                directive @deprecated(reason: String!) on FIELD
                """,
                [
                    {
                        "message": "Directive '@deprecated' argument 'reason' of type"
                        " 'String!' is required, but it was not provided.",
                        "locations": [(3, 31)],
                    },
                ],
            )

        def missing_arg_on_directive_defined_in_schema_extension():
            schema = build_schema(
                """
                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_errors(
                """
                directive @test(arg: String!) on OBJECT

                extend type Query  @test
                """,
                [
                    {
                        "message": "Directive '@test' argument 'arg' of type"
                        " 'String!' is required, but it was not provided.",
                        "locations": [(4, 36)],
                    },
                ],
                schema,
            )

        def missing_arg_on_directive_used_in_schema_extension():
            schema = build_schema(
                """
                directive @test(arg: String!) on OBJECT

                type Query {
                  foo: String
                }
                """
            )
            assert_sdl_errors(
                """
                extend type Query  @test
                """,
                [
                    {
                        "message": "Directive '@test' argument 'arg' of type"
                        " 'String!' is required, but it was not provided.",
                        "locations": [(2, 36)],
                    },
                ],
                schema,
            )



from math import nan, inf
from contextlib import contextmanager
from importlib import import_module
from typing import Any, Dict, FrozenSet, List, Set, Tuple

from pytest import mark

from graphql.pyutils import inspect, Undefined
from graphql.type import (
    GraphQLDirective,
    GraphQLField,
    GraphQLInt,
    GraphQLList,
    GraphQLObjectType,
    GraphQLNonNull,
    GraphQLString,
)

inspect_module = import_module(inspect.__module__)


@contextmanager
def increased_recursive_depth():
    inspect_module.max_recursive_depth += 1  # type: ignore
    try:
        yield inspect
    finally:
        inspect_module.max_recursive_depth -= 1  # type: ignore


@contextmanager
def increased_str_size():
    inspect_module.max_str_size *= 2  # type: ignore
    try:
        yield inspect
    finally:
        inspect_module.max_str_size //= 2  # type: ignore


@contextmanager
def increased_list_size():
    inspect_module.max_list_size *= 2  # type: ignore
    try:
        yield inspect
    finally:
        inspect_module.max_list_size //= 2  # type: ignore


def describe_inspect():
    def inspect_invalid():
        assert inspect(Undefined) == "Undefined"

    def inspect_none():
        assert inspect(None) == "None"

    def inspect_boolean():
        assert inspect(True) == "True"
        assert inspect(False) == "False"

    def inspect_string():
        for s in "", "abc", "foo\tbar \u265e\0", "'", "'":
            assert inspect(s) == repr(s)

    def overly_large_string():
        s = "foo" * 100
        r = repr(s)
        assert inspect(s) == r[:118] + "..." + r[-119:]
        with increased_str_size():
            assert inspect(s) == r

    def inspect_bytes():
        for b in b"", b"abc", b"foo\tbar \x7f\xff\0", b"'", b"'":
            assert inspect(b) == repr(b)
            a = bytearray(b)
            assert inspect(a) == repr(a)

    def overly_many_bytes():
        b = b"foo" * 100
        r = repr(b)
        assert inspect(b) == r[:118] + "..." + r[-119:]
        a = bytearray(b)
        r = repr(a)
        assert inspect(a) == r[:118] + "..." + r[-119:]

    def inspect_numbers():
        assert inspect(0) == "0"
        assert inspect(0.0) == "0.0"
        assert inspect(314) == "314"
        assert inspect(3.14) == "3.14"
        assert inspect(complex(1, 2)) == "(1+2j)"
        assert inspect(nan) == "nan"
        assert inspect(inf) == "inf"
        assert inspect(-inf) == "-inf"

    def overly_large_int():
        n = int("123" * 100)
        r = repr(n)
        assert inspect(n) == r[:118] + "..." + r[-119:]
        with increased_str_size():
            assert inspect(n) == r

    def inspect_function():
        assert inspect(lambda: 0) == "<function>"  # pragma: no cover

        def test_func():
            pass

        assert inspect(test_func) == "<function test_func>"

    def inspect_exception():
        assert inspect(ValueError) == "<exception class ValueError>"
        assert inspect(ArithmeticError(42)) == "<exception ArithmeticError>"

    def inspect_class_and_method():
        class TestClass:
            def test_method(self):
                pass

        assert inspect(TestClass) == "<class TestClass>"
        assert inspect(TestClass()) == "<TestClass instance>"
        assert inspect(TestClass.test_method) == "<function test_method>"
        assert inspect(TestClass().test_method) == "<method test_method>"

    def inspect_unknown_object():
        class MetaClass(type):
            __name__ = ""

        class TestClass(metaclass=MetaClass):
            pass

        assert inspect(TestClass()) == "<object>"

    def inspect_generator():
        def test_generator():
            yield None  # pragma: no cover

        assert inspect(test_generator) == "<generator function test_generator>"
        assert inspect(test_generator()) == "<generator test_generator>"

    @mark.asyncio
    async def inspect_coroutine():
        async def test_coroutine():
            pass

        assert inspect(test_coroutine) == "<coroutine function test_coroutine>"
        coroutine_object = test_coroutine()
        assert inspect(coroutine_object) == "<coroutine test_coroutine>"
        await coroutine_object  # avoid warning

    def inspect_async_generator():
        async def test_async_generator():
            yield None  # pragma: no cover

        assert inspect(test_async_generator) == (
            "<async generator function test_async_generator>"
        )
        assert inspect(test_async_generator()) == (
            "<async generator test_async_generator>"
        )

    def inspect_lists():
        assert inspect([]) == "[]"
        assert inspect([None]) == "[None]"
        assert inspect([[None]]) == "[[None]]"
        assert inspect([1, nan]) == "[1, nan]"
        assert inspect([["a", "b"], "c"]) == "[['a', 'b'], 'c']"

    def inspect_overly_large_list():
        s: List[int] = list(range(20))
        assert inspect(s) == "[0, 1, 2, 3, 4, ..., 16, 17, 18, 19]"
        with increased_list_size():
            assert inspect(s) == repr(s)

    def inspect_overly_nested_list():
        s: List[List[List]] = [[[]]]
        assert inspect(s) == "[[[]]]"
        s = [[[1, 2, 3]]]
        assert inspect(s) == "[[[...]]]"
        with increased_recursive_depth():
            assert inspect(s) == repr(s)

    def inspect_recursive_list():
        s: List[Any] = [1, 2, 3]
        s[1] = s
        assert inspect(s) == "[1, [...], 3]"

    def inspect_tuples():
        assert inspect(()) == "()"
        assert inspect((None,)) == "(None,)"
        assert inspect(((None,),)) == "((None,),)"
        assert inspect((1, nan)) == "(1, nan)"
        assert inspect((("a", "b"), "c")) == "(('a', 'b'), 'c')"

    def inspect_overly_large_tuple():
        s = tuple(range(20))
        assert inspect(s) == "(0, 1, 2, 3, 4, ..., 16, 17, 18, 19)"
        with increased_list_size():
            assert inspect(s) == repr(s)

    def inspect_overly_nested_tuple():
        s: Tuple[Tuple[Tuple]] = (((),),)
        assert inspect(s) == "(((),),)"
        s = (((1, 2, 3),),)
        assert inspect(s) == "(((...),),)"
        with increased_recursive_depth():
            assert inspect(s) == repr(s)

    def inspect_recursive_tuple():
        s: List[Any] = [1, 2, 3]
        s[1] = s
        t = tuple(s)
        assert inspect(t) == "(1, [1, [...], 3], 3)"

    def mixed_lists_and_tuples():
        assert inspect(["a", ("b",)]) == "['a', ('b',)]"

    def mixed_lists_and_tuples_with_various_objects():
        class TestClass:
            pass

        assert inspect([TestClass, (TestClass,), ValueError()]) == (
            "[<class TestClass>, (<class TestClass>,), <exception ValueError>]"
        )

    def inspect_dicts():
        assert inspect({}) == "{}"
        assert inspect({"a": 1}) == "{'a': 1}"
        assert inspect({"a": 1, "b": 2}) == "{'a': 1, 'b': 2}"
        assert inspect({"list": [None, 0]}) == "{'list': [None, 0]}"
        assert inspect({"a": True, "b": None}) == "{'a': True, 'b': None}"

    def inspect_overly_large_dict():
        s = dict(zip((chr(97 + i) for i in range(20)), range(20)))
        assert (
            inspect(s) == "{'a': 0, 'b': 1, 'c': 2, 'd': 3, 'e': 4,"
            " ..., 'q': 16, 'r': 17, 's': 18, 't': 19}"
        )
        with increased_list_size():
            assert inspect(s) == repr(s)

    def inspect_overly_nested_dict():
        s: Dict[str, Dict[str, Dict]] = {"a": {"b": {}}}
        assert inspect(s) == "{'a': {'b': {}}}"
        s = {"a": {"b": {"c": 3}}}
        assert inspect(s) == "{'a': {'b': {...}}}"
        with increased_recursive_depth():
            assert inspect(s) == repr(s)

    def inspect_recursive_dict():
        s: Dict[int, Any] = {}
        s[1] = s
        assert inspect(s) == "{1: {...}}"

    def inspect_sets():
        assert inspect(set()) == "set()"
        assert inspect({"a"}) == "{'a'}"
        assert inspect({"a", 1}) in ("{'a', 1}", "{1, 'a'}")  # sets are unordered

    def inspect_overly_large_set():
        s = set(range(20))
        r = inspect(s)
        assert r.startswith("{") and r.endswith("}")
        assert "..., " in r and "5" not in s  # sets are unordered
        assert len(r) == 36
        with increased_list_size():
            assert inspect(s) == repr(s)

    def inspect_overly_nested_set():
        s: List[List[Set]] = [[set()]]
        assert inspect(s) == "[[set()]]"
        s = [[{1, 2, 3}]]
        assert inspect(s) == "[[set(...)]]"
        with increased_recursive_depth():
            assert inspect(s) == repr(s)

    def inspect_frozensets():
        assert inspect(frozenset()) == "frozenset()"
        assert inspect(frozenset(["a"])) == "frozenset({'a'})"
        assert inspect(frozenset(["a", 1])) in (
            "frozenset({'a', 1})",
            "frozenset({1, 'a'})",
        )  # frozensets are unordered

    def inspect_overly_large_frozenset():
        s = frozenset(range(20))
        r = inspect(s)
        assert r.startswith("frozenset({") and r.endswith("})")
        assert "..., " in r and "5" not in s  # frozensets are unordered
        assert len(r) == 47
        with increased_list_size():
            assert inspect(s) == repr(s)

    def inspect_overly_nested_frozenset():
        s: FrozenSet[FrozenSet[FrozenSet]] = frozenset([frozenset([frozenset()])])
        assert inspect(s) == "frozenset({frozenset({frozenset()})})"
        s = frozenset([frozenset([frozenset([1, 2, 3])])])
        assert inspect(s) == "frozenset({frozenset({frozenset(...)})})"
        with increased_recursive_depth():
            assert inspect(s) == repr(s)

    def mixed_recursive_dict_and_list():
        s: Any = {1: []}
        s[1].append(s)
        assert inspect(s) == "{1: [{...}]}"
        s = [1, 2, 3]
        s[1] = {2: s}
        assert inspect(s) == "[1, {2: [...]}, 3]"

    def mixed_dicts_and_sets():
        assert inspect({"a": {"b"}}) == "{'a': {'b'}}"
        assert inspect({1: [], 2: (), 3: set()}) == "{1: [], 2: (), 3: set()}"
        assert inspect([(set(),), {None: {()}}]) == "[(set(),), {None: set(...)}]"

    def mixed_dicts_and_sets_with_various_objects():
        class TestClass:
            pass

        assert inspect({TestClass: {ValueError()}, ValueError: {TestClass()}}) == (
            "{<class TestClass>: {<exception ValueError>},"
            " <exception class ValueError>: {<TestClass instance>}}"
        )

    def inspect_graphql_types():
        assert inspect(GraphQLInt) == "Int"
        assert inspect(GraphQLString) == "String"
        assert inspect(GraphQLNonNull(GraphQLString)) == "String!"
        assert inspect(GraphQLList(GraphQLString)) == "[String]"
        test_object_type = GraphQLObjectType(
            "TestObjectType", {"test": GraphQLField(GraphQLString)}
        )
        assert inspect(test_object_type) == "TestObjectType"
        test_directive = GraphQLDirective("TestDirective", [])
        assert inspect(test_directive) == "@TestDirective"

    def custom_inspect():
        class TestClass:
            @staticmethod
            def __inspect__():
                return "<custom magic method inspect>"

        assert inspect(TestClass()) == "<custom magic method inspect>"

    def custom_inspect_that_uses_self():
        class TestClass:
            str = "Hello World!"

            def __inspect__(self):
                return self.str

        assert inspect(TestClass()) == "Hello World!"

    def custom_inspect_that_returns_a_list():
        class TestClass:
            @staticmethod
            def __inspect__():
                return [1, 2, 3]

        assert inspect(TestClass()) == "[1, 2, 3]"

    def custom_inspect_that_returns_an_overly_large_string():
        s = "foo" * 100

        class TestClass:
            @staticmethod
            def __inspect__():
                return s

        value = TestClass()

        assert inspect(value) == s[:118] + "..." + s[-119:]
        with increased_str_size():
            assert inspect(value) == s

    def custom_inspect_that_is_recursive():
        class TestClass:
            def __inspect__(self):
                return self

        assert inspect(TestClass()) == "<TestClass instance>"



from graphql.pyutils import Undefined


def describe_invalid():
    def has_repr():
        assert repr(Undefined) == "Undefined"

    def has_str():
        assert str(Undefined) == "Undefined"

    def is_hashable():
        assert hash(Undefined) == hash(Undefined)
        assert hash(Undefined) != hash(None)
        assert hash(Undefined) != hash(False)
        assert hash(Undefined) != hash(True)

    def as_bool_is_false():
        assert bool(Undefined) is False

    def only_equal_to_itself():
        assert Undefined == Undefined
        assert not Undefined != Undefined
        none_object = None
        assert Undefined != none_object
        assert not Undefined == none_object
        false_object = False
        assert Undefined != false_object
        assert not Undefined == false_object



from typing import Any, cast

from ....error import GraphQLError
from ....language import ArgumentNode, EnumValueNode, FieldNode, ObjectFieldNode
from ....type import GraphQLInputObjectType, get_named_type, is_input_object_type
from .. import ValidationRule

__all__ = ["NoDeprecatedCustomRule"]


class NoDeprecatedCustomRule(ValidationRule):
    """No deprecated

    A GraphQL document is only valid if all selected fields and all used enum values
    have not been deprecated.

    Note: This rule is optional and is not part of the Validation section of the GraphQL
    Specification. The main purpose of this rule is detection of deprecated usages and
    not necessarily to forbid their use when querying a service.
    """

    def enter_field(self, node: FieldNode, *_args: Any) -> None:
        context = self.context
        field_def = context.get_field_def()
        if field_def:
            deprecation_reason = field_def.deprecation_reason
            if deprecation_reason is not None:
                parent_type = context.get_parent_type()
                parent_name = parent_type.name  # type: ignore
                self.report_error(
                    GraphQLError(
                        f"The field {parent_name}.{node.name.value}"
                        f" is deprecated. {deprecation_reason}",
                        node,
                    )
                )

    def enter_argument(self, node: ArgumentNode, *_args: Any) -> None:
        context = self.context
        arg_def = context.get_argument()
        if arg_def:
            deprecation_reason = arg_def.deprecation_reason
            if deprecation_reason is not None:
                directive_def = context.get_directive()
                arg_name = node.name.value
                if directive_def is None:
                    parent_type = context.get_parent_type()
                    parent_name = parent_type.name  # type: ignore
                    field_def = context.get_field_def()
                    field_name = field_def.ast_node.name.value  # type: ignore
                    self.report_error(
                        GraphQLError(
                            f"Field '{parent_name}.{field_name}' argument"
                            f" '{arg_name}' is deprecated. {deprecation_reason}",
                            node,
                        )
                    )
                else:
                    self.report_error(
                        GraphQLError(
                            f"Directive '@{directive_def.name}' argument"
                            f" '{arg_name}' is deprecated. {deprecation_reason}",
                            node,
                        )
                    )

    def enter_object_field(self, node: ObjectFieldNode, *_args: Any) -> None:
        context = self.context
        input_object_def = get_named_type(context.get_parent_input_type())
        if is_input_object_type(input_object_def):
            input_field_def = cast(GraphQLInputObjectType, input_object_def).fields.get(
                node.name.value
            )
            if input_field_def:
                deprecation_reason = input_field_def.deprecation_reason
                if deprecation_reason is not None:
                    field_name = node.name.value
                    input_object_name = input_object_def.name  # type: ignore
                    self.report_error(
                        GraphQLError(
                            f"The input field {input_object_name}.{field_name}"
                            f" is deprecated. {deprecation_reason}",
                            node,
                        )
                    )

    def enter_enum_value(self, node: EnumValueNode, *_args: Any) -> None:
        context = self.context
        enum_value_def = context.get_enum_value()
        if enum_value_def:
            deprecation_reason = enum_value_def.deprecation_reason
            if deprecation_reason is not None:  # pragma: no cover else
                enum_type_def = get_named_type(context.get_input_type())
                enum_type_name = enum_type_def.name  # type: ignore
                self.report_error(
                    GraphQLError(
                        f"The enum value '{enum_type_name}.{node.value}'"
                        f" is deprecated. {deprecation_reason}",
                        node,
                    )
                )



from itertools import chain
from typing import Collection

from ..language.ast import DocumentNode

__all__ = ["concat_ast"]


def concat_ast(asts: Collection[DocumentNode]) -> DocumentNode:
    """Concat ASTs.

    Provided a collection of ASTs, presumably each from different files, concatenate
    the ASTs together into batched AST, useful for validating many GraphQL source files
    which together represent one conceptual application.
    """
    return DocumentNode(
        definitions=list(chain.from_iterable(document.definitions for document in asts))
    )



from operator import attrgetter
from typing import Any, List, Union

from pytest import mark, raises

from graphql.language import parse, DirectiveLocation
from graphql.pyutils import inspect
from graphql.type import (
    assert_directive,
    assert_enum_type,
    assert_input_object_type,
    assert_interface_type,
    assert_object_type,
    assert_scalar_type,
    assert_union_type,
    assert_valid_schema,
    is_input_type,
    is_output_type,
    validate_schema,
    GraphQLArgument,
    GraphQLDirective,
    GraphQLEnumType,
    GraphQLField,
    GraphQLInputField,
    GraphQLInputType,
    GraphQLInputObjectType,
    GraphQLInt,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLNamedType,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLOutputType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
)
from graphql.utilities import build_schema, extend_schema

from ..utils import dedent

SomeSchema = build_schema(
    """
    scalar SomeScalar

    interface SomeInterface { f: SomeObject }

    type SomeObject implements SomeInterface { f: SomeObject }

    union SomeUnion = SomeObject

    enum SomeEnum { ONLY }

    input SomeInputObject { val: String = "hello" }

    directive @SomeDirective on QUERY
    """
)

get_type = SomeSchema.get_type
SomeScalarType = assert_scalar_type(get_type("SomeScalar"))
SomeInterfaceType = assert_interface_type(get_type("SomeInterface"))
SomeObjectType = assert_object_type(get_type("SomeObject"))
SomeUnionType = assert_union_type(get_type("SomeUnion"))
SomeEnumType = assert_enum_type(get_type("SomeEnum"))
SomeInputObjectType = assert_input_object_type(get_type("SomeInputObject"))
SomeDirective = assert_directive(SomeSchema.get_directive("SomeDirective"))


def with_modifiers(
    type_: GraphQLNamedType,
) -> List[Union[GraphQLNamedType, GraphQLNonNull, GraphQLList]]:
    return [
        type_,
        GraphQLList(type_),
        GraphQLNonNull(type_),
        GraphQLNonNull(GraphQLList(type_)),
    ]


output_types = [
    *with_modifiers(GraphQLString),
    *with_modifiers(SomeScalarType),
    *with_modifiers(SomeEnumType),
    *with_modifiers(SomeObjectType),
    *with_modifiers(SomeUnionType),
    *with_modifiers(SomeInterfaceType),
]

not_output_types = with_modifiers(SomeInputObjectType)

input_types = [
    *with_modifiers(GraphQLString),
    *with_modifiers(SomeScalarType),
    *with_modifiers(SomeEnumType),
    *with_modifiers(SomeInputObjectType),
]

not_input_types = [
    *with_modifiers(SomeObjectType),
    *with_modifiers(SomeUnionType),
    *with_modifiers(SomeInterfaceType),
]

not_graphql_types = [
    type("IntType", (int,), {"name": "IntType"}),
    type("FloatType", (float,), {"name": "FloatType"}),
    type("StringType", (str,), {"name": "StringType"}),
]


get_name = attrgetter("__class__.__name__")


def schema_with_field_type(type_):
    return GraphQLSchema(
        query=GraphQLObjectType(name="Query", fields={"f": GraphQLField(type_)})
    )


def describe_type_system_a_schema_must_have_object_root_types():
    def accepts_a_schema_whose_query_type_is_an_object_type():
        schema = build_schema(
            """
            type Query {
              test: String
            }
            """
        )
        assert validate_schema(schema) == []

        schema_with_def = build_schema(
            """
            schema {
              query: QueryRoot
            }

            type QueryRoot {
              test: String
            }
            """
        )

        assert validate_schema(schema_with_def) == []

    def accepts_a_schema_whose_query_and_mutation_types_are_object_types():
        schema = build_schema(
            """
            type Query {
              test: String
            }

            type Mutation {
              test: String
            }
            """
        )
        assert validate_schema(schema) == []

        schema_with_def = build_schema(
            """
            schema {
              query: QueryRoot
              mutation: MutationRoot
            }

            type QueryRoot {
              test: String
            }

            type MutationRoot {
              test: String
            }
            """
        )
        assert validate_schema(schema_with_def) == []

    def accepts_a_schema_whose_query_and_subscription_types_are_object_types():
        schema = build_schema(
            """
            type Query {
              test: String
            }

            type Subscription {
              test: String
            }
            """
        )
        assert validate_schema(schema) == []

        schema_with_def = build_schema(
            """
            schema {
              query: QueryRoot
              subscription: SubscriptionRoot
            }

            type QueryRoot {
              test: String
            }

            type SubscriptionRoot {
              test: String
            }
            """
        )
        assert validate_schema(schema_with_def) == []

    def rejects_a_schema_without_a_query_type():
        schema = build_schema(
            """
            type Mutation {
              test: String
            }
            """
        )
        assert validate_schema(schema) == [
            {"message": "Query root type must be provided.", "locations": None}
        ]

        schema_with_def = build_schema(
            """
            schema {
              mutation: MutationRoot
            }

            type MutationRoot {
              test: String
            }
            """
        )
        assert validate_schema(schema_with_def) == [
            {"message": "Query root type must be provided.", "locations": [(2, 13)]}
        ]

    def rejects_a_schema_whose_query_root_type_is_not_an_object_type():
        schema = build_schema(
            """
            input Query {
              test: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Query root type must be Object type,"
                " it cannot be Query.",
                "locations": [(2, 13)],
            }
        ]

        schema_with_def = build_schema(
            """
            schema {
              query: SomeInputObject
            }

            input SomeInputObject {
              test: String
            }
            """
        )
        assert validate_schema(schema_with_def) == [
            {
                "message": "Query root type must be Object type,"
                " it cannot be SomeInputObject.",
                "locations": [(3, 22)],
            }
        ]

    def rejects_a_schema_whose_mutation_type_is_an_input_type():
        schema = build_schema(
            """
            type Query {
              field: String
            }

            input Mutation {
              test: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Mutation root type must be Object type if provided,"
                " it cannot be Mutation.",
                "locations": [(6, 13)],
            }
        ]

        schema_with_def = build_schema(
            """
            schema {
              query: Query
              mutation: SomeInputObject
            }

            type Query {
              field: String
            }

            input SomeInputObject {
              test: String
            }
            """
        )
        assert validate_schema(schema_with_def) == [
            {
                "message": "Mutation root type must be Object type if provided,"
                " it cannot be SomeInputObject.",
                "locations": [(4, 25)],
            }
        ]

    def rejects_a_schema_whose_subscription_type_is_an_input_type():
        schema = build_schema(
            """
            type Query {
              field: String
            }

            input Subscription {
              test: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Subscription root type must be Object type if"
                " provided, it cannot be Subscription.",
                "locations": [(6, 13)],
            }
        ]

        schema_with_def = build_schema(
            """
            schema {
              query: Query
              subscription: SomeInputObject
            }

            type Query {
              field: String
            }

            input SomeInputObject {
              test: String
            }
            """
        )
        assert validate_schema(schema_with_def) == [
            {
                "message": "Subscription root type must be Object type if"
                " provided, it cannot be SomeInputObject.",
                "locations": [(4, 29)],
            }
        ]

    def rejects_a_schema_extended_with_invalid_root_types():
        schema = build_schema(
            """
            input SomeInputObject {
              test: String
            }
            """
        )
        schema = extend_schema(
            schema,
            parse(
                """
                extend schema {
                  query: SomeInputObject
                }
                """
            ),
        )
        schema = extend_schema(
            schema,
            parse(
                """
                extend schema {
                  mutation: SomeInputObject
                }
                """
            ),
        )
        schema = extend_schema(
            schema,
            parse(
                """
                extend schema {
                  subscription: SomeInputObject
                }
                """
            ),
        )
        assert validate_schema(schema) == [
            {
                "message": "Query root type must be Object type,"
                " it cannot be SomeInputObject.",
                "locations": [(3, 26)],
            },
            {
                "message": "Mutation root type must be Object type"
                " if provided, it cannot be SomeInputObject.",
                "locations": [(3, 29)],
            },
            {
                "message": "Subscription root type must be Object type"
                " if provided, it cannot be SomeInputObject.",
                "locations": [(3, 33)],
            },
        ]

    def rejects_a_schema_whose_types_are_incorrectly_type():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            # noinspection PyTypeChecker
            GraphQLSchema(
                SomeObjectType,
                types=[{"name": "SomeType"}, SomeDirective],  # type: ignore
            )
        assert str(exc_info.value) == (
            "Schema types must be specified as a collection of GraphQL types."
        )
        # construct invalid schema manually
        schema = GraphQLSchema(SomeObjectType)
        schema.type_map = {
            "SomeType": {"name": "SomeType"},  # type: ignore
            "SomeDirective": SomeDirective,  # type: ignore
        }
        assert validate_schema(schema) == [
            {"message": "Expected GraphQL named type but got: {'name': 'SomeType'}."},
            {"message": "Expected GraphQL named type but got: @SomeDirective."},
        ]

    def rejects_a_schema_whose_directives_are_incorrectly_typed():
        schema = GraphQLSchema(
            SomeObjectType,
            directives=[None, "SomeDirective", SomeScalarType],  # type: ignore
        )
        assert validate_schema(schema) == [
            {"message": "Expected directive but got: None."},
            {"message": "Expected directive but got: 'SomeDirective'."},
            {"message": "Expected directive but got: SomeScalar."},
        ]


def describe_type_system_objects_must_have_fields():
    def accepts_an_object_type_with_fields_object():
        schema = build_schema(
            """
            type Query {
              field: SomeObject
            }

            type SomeObject {
              field: String
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_object_type_with_missing_fields():
        schema = build_schema(
            """
            type Query {
              test: IncompleteObject
            }

            type IncompleteObject
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Type IncompleteObject must define one or more fields.",
                "locations": [(6, 13)],
            }
        ]

        manual_schema = schema_with_field_type(
            GraphQLObjectType("IncompleteObject", {})
        )
        msg = validate_schema(manual_schema)[0].message
        assert msg == "Type IncompleteObject must define one or more fields."

        manual_schema_2 = schema_with_field_type(
            GraphQLObjectType("IncompleteObject", lambda: {})
        )
        msg = validate_schema(manual_schema_2)[0].message
        assert msg == "Type IncompleteObject must define one or more fields."

    def rejects_an_object_type_with_incorrectly_named_fields():
        schema = schema_with_field_type(
            GraphQLObjectType("SomeObject", {"__badName": GraphQLField(GraphQLString)})
        )
        msg = validate_schema(schema)[0].message
        assert msg == (
            "Name '__badName' must not begin with '__',"
            " which is reserved by GraphQL introspection."
        )


def describe_type_system_field_args_must_be_properly_named():
    def accepts_field_args_with_valid_names():
        schema = schema_with_field_type(
            GraphQLObjectType(
                "SomeObject",
                {
                    "goodField": GraphQLField(
                        GraphQLString, args={"goodArg": GraphQLArgument(GraphQLString)}
                    )
                },
            )
        )
        assert validate_schema(schema) == []

    def rejects_field_args_with_invalid_names():
        schema = schema_with_field_type(
            GraphQLObjectType(
                "SomeObject",
                {
                    "badField": GraphQLField(
                        GraphQLString,
                        args={"__badName": GraphQLArgument(GraphQLString)},
                    )
                },
            )
        )

        msg = validate_schema(schema)[0].message
        assert msg == (
            "Name '__badName' must not begin with '__',"
            " which is reserved by GraphQL introspection."
        )


def describe_type_system_union_types_must_be_valid():
    def accepts_a_union_type_with_member_types():
        schema = build_schema(
            """
            type Query {
              test: GoodUnion
            }

            type TypeA {
              field: String
            }

            type TypeB {
              field: String
            }

            union GoodUnion =
              | TypeA
              | TypeB
            """
        )
        assert validate_schema(schema) == []

    def rejects_a_union_type_with_empty_types():
        schema = build_schema(
            """
            type Query {
              test: BadUnion
            }

            union BadUnion
            """
        )

        schema = extend_schema(
            schema,
            parse(
                """
                directive @test on UNION

                extend union BadUnion @test
                """
            ),
        )

        assert validate_schema(schema) == [
            {
                "message": "Union type BadUnion must define one or more member types.",
                "locations": [(6, 13), (4, 17)],
            }
        ]

    def rejects_a_union_type_with_duplicated_member_type():
        schema = build_schema(
            """
            type Query {
              test: BadUnion
            }

            type TypeA {
              field: String
            }

            type TypeB {
              field: String
            }

            union BadUnion =
              | TypeA
              | TypeB
              | TypeA
            """
        )

        assert validate_schema(schema) == [
            {
                "message": "Union type BadUnion can only include type TypeA once.",
                "locations": [(15, 17), (17, 17)],
            }
        ]

        schema = extend_schema(schema, parse("extend union BadUnion = TypeB"))

        assert validate_schema(schema) == [
            {
                "message": "Union type BadUnion can only include type TypeA once.",
                "locations": [(15, 17), (17, 17)],
            },
            {
                "message": "Union type BadUnion can only include type TypeB once.",
                "locations": [(16, 17), (1, 25)],
            },
        ]

    def rejects_a_union_type_with_non_object_member_types():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  test: BadUnion
                }

                type TypeA {
                  field: String
                }

                type TypeB {
                  field: String
                }

                union BadUnion =
                  | TypeA
                  | String
                  | TypeB
                """
            )
        assert str(exc_info.value) == (
            "BadUnion types must be specified"
            " as a collection of GraphQLObjectType instances."
        )
        # construct invalid schema manually
        schema = build_schema(
            """
            type Query {
              test: BadUnion
            }

            type TypeA {
              field: String
            }

            type TypeB {
              field: String
            }

            union BadUnion =
              | TypeA
              | TypeA
              | TypeB
            """
        )
        with raises(TypeError) as exc_info:
            extend_schema(schema, parse("extend union BadUnion = Int"))
        assert str(exc_info.value) == (
            "BadUnion types must be specified"
            " as a collection of GraphQLObjectType instances."
        )
        schema = extend_schema(schema, parse("extend union BadUnion = TypeB"))
        bad_union: Any = schema.get_type("BadUnion")
        types = bad_union.types
        assert isinstance(types, tuple)
        types = list(types)
        assert types[1].name == "TypeA"
        types[1] = GraphQLString
        assert types[3].name == "TypeB"
        types[3] = GraphQLInt
        bad_union.types = tuple(types)
        bad_union.ast_node.types[1].name.value = "String"
        bad_union.extension_ast_nodes[0].types[0].name.value = "Int"
        assert validate_schema(schema) == [
            {
                "message": "Union type BadUnion can only include Object types,"
                " it cannot include String.",
                "locations": [(16, 17)],
            },
            {
                "message": "Union type BadUnion can only include Object types,"
                " it cannot include Int.",
                "locations": [(1, 25)],
            },
        ]

        bad_union_member_types = [
            GraphQLString,
            GraphQLNonNull(SomeObjectType),
            GraphQLList(SomeObjectType),
            SomeInterfaceType,
            SomeUnionType,
            SomeEnumType,
            SomeInputObjectType,
        ]
        for member_type in bad_union_member_types:
            # invalid union type cannot be built with Python
            bad_union = GraphQLUnionType(
                "BadUnion", types=[member_type]  # type: ignore
            )
            with raises(TypeError) as exc_info:
                schema_with_field_type(bad_union)
            assert str(exc_info.value) == (
                "BadUnion types must be specified"
                " as a collection of GraphQLObjectType instances."
            )
            # noinspection PyPropertyAccess
            bad_union.types = []
            bad_schema = schema_with_field_type(bad_union)
            # noinspection PyPropertyAccess
            bad_union.types = [member_type]
            assert validate_schema(bad_schema) == [
                {
                    "message": "Union type BadUnion can only include Object types,"
                    + f" it cannot include {inspect(member_type)}."
                }
            ]


def describe_type_system_input_objects_must_have_fields():
    def accepts_an_input_object_type_with_fields():
        schema = build_schema(
            """
            type Query {
               field(arg: SomeInputObject): String
            }

            input SomeInputObject {
              field: String
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_input_object_type_with_missing_fields():
        schema = build_schema(
            """
            type Query {
              field(arg: SomeInputObject): String
            }

            input SomeInputObject
            """
        )
        schema = extend_schema(
            schema,
            parse(
                """
                directive @test on INPUT_OBJECT

                extend input SomeInputObject @test
                """
            ),
        )
        assert validate_schema(schema) == [
            {
                "message": "Input Object type SomeInputObject"
                " must define one or more fields.",
                "locations": [(6, 13), (4, 17)],
            }
        ]

    def accepts_an_input_object_with_breakable_circular_reference():
        schema = build_schema(
            """
            type Query {
              field(arg: SomeInputObject): String
            }

            input SomeInputObject {
              self: SomeInputObject
              arrayOfSelf: [SomeInputObject]
              nonNullArrayOfSelf: [SomeInputObject]!
              nonNullArrayOfNonNullSelf: [SomeInputObject!]!
              intermediateSelf: AnotherInputObject
            }

            input AnotherInputObject {
              parent: SomeInputObject
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_input_object_with_non_breakable_circular_reference():
        schema = build_schema(
            """
            type Query {
              field(arg: SomeInputObject): String
            }

            input SomeInputObject {
              startLoop: AnotherInputObject!
            }

            input AnotherInputObject {
              nextInLoop: YetAnotherInputObject!
            }

            input YetAnotherInputObject {
              closeLoop: SomeInputObject!
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Cannot reference Input Object 'SomeInputObject'"
                " within itself through a series of non-null fields:"
                " 'startLoop.nextInLoop.closeLoop'.",
                "locations": [(7, 15), (11, 15), (15, 15)],
            }
        ]

    def rejects_an_input_object_with_multiple_non_breakable_circular_reference():
        schema = build_schema(
            """
            type Query {
              field(arg: SomeInputObject): String
            }

            input SomeInputObject {
              startLoop: AnotherInputObject!
            }

            input AnotherInputObject {
              closeLoop: SomeInputObject!
              startSecondLoop: YetAnotherInputObject!
            }

            input YetAnotherInputObject {
              closeSecondLoop: AnotherInputObject!
              nonNullSelf: YetAnotherInputObject!
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Cannot reference Input Object 'SomeInputObject'"
                " within itself through a series of non-null fields:"
                " 'startLoop.closeLoop'.",
                "locations": [(7, 15), (11, 15)],
            },
            {
                "message": "Cannot reference Input Object 'AnotherInputObject'"
                " within itself through a series of non-null fields:"
                " 'startSecondLoop.closeSecondLoop'.",
                "locations": [(12, 15), (16, 15)],
            },
            {
                "message": "Cannot reference Input Object 'YetAnotherInputObject'"
                " within itself through a series of non-null fields:"
                " 'nonNullSelf'.",
                "locations": [(17, 15)],
            },
        ]

    def rejects_an_input_object_type_with_incorrectly_typed_fields():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  field(arg: SomeInputObject): String
                }

                type SomeObject {
                  field: String
                }

                union SomeUnion = SomeObject

                input SomeInputObject {
                  badObject: SomeObject
                  badUnion: SomeUnion
                  goodInputObject: SomeInputObject
                }
                """
            )
        assert str(exc_info.value) == (
            "SomeInputObject fields cannot be resolved."
            " Input field type must be a GraphQL input type."
        )
        # construct invalid schema manually
        schema = build_schema(
            """
            type Query {
              field(arg: SomeInputObject): String
            }

            type SomeObject {
              field: String
            }

            union SomeUnion = SomeObject

            input SomeInputObject {
              badObject: SomeInputObject
              badUnion: SomeInputObject
              goodInputObject: SomeInputObject
            }
            """
        )
        some_input_obj: Any = schema.get_type("SomeInputObject")
        some_input_obj.fields["badObject"].type = schema.get_type("SomeObject")
        some_input_obj.fields["badUnion"].type = schema.get_type("SomeUnion")
        assert validate_schema(schema) == [
            {
                "message": "The type of SomeInputObject.badObject must be Input Type"
                " but got: SomeObject.",
                "locations": [(13, 26)],
            },
            {
                "message": "The type of SomeInputObject.badUnion must be Input Type"
                " but got: SomeUnion.",
                "locations": [(14, 25)],
            },
        ]

    def rejects_an_input_object_type_with_required_arguments_that_is_deprecated():
        schema = build_schema(
            """
            type Query {
              field(arg: SomeInputObject): String
            }

            input SomeInputObject {
              badField: String! @deprecated
              optionalField: String @deprecated
              anotherOptionalField: String! = "" @deprecated
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Required input field SomeInputObject.badField"
                " cannot be deprecated.",
                "locations": [(7, 33), (7, 25)],
            }
        ]


def describe_type_system_enum_types_must_be_well_defined():
    def rejects_an_enum_type_without_values():
        schema = build_schema(
            """
            type Query {
              field: SomeEnum
            }

            enum SomeEnum
            """
        )

        schema = extend_schema(
            schema,
            parse(
                """
                directive @test on ENUM

                extend enum SomeEnum @test
                """
            ),
        )

        assert validate_schema(schema) == [
            {
                "message": "Enum type SomeEnum must define one or more values.",
                "locations": [(6, 13), (4, 17)],
            }
        ]

    def rejects_an_enum_type_with_incorrectly_named_values():
        schema = schema_with_field_type(
            GraphQLEnumType("SomeEnum", values={"__badName": {}})
        )

        assert validate_schema(schema) == [
            {
                "message": "Name '__badName' must not begin with '__',"
                " which is reserved by GraphQL introspection."
            }
        ]


def describe_type_system_object_fields_must_have_output_types():
    def _schema_with_object_field(type_: GraphQLOutputType) -> GraphQLSchema:
        if is_output_type(type_):
            field = GraphQLField(type_)
        else:
            # invalid field cannot be built with Python directly
            with raises(TypeError) as exc_info:
                GraphQLField(type_)
            assert str(exc_info.value) == "Field type must be an output type."
            # therefore we need to monkey-patch a valid field
            field = GraphQLField(GraphQLString)
            field.type = type_
        bad_object_type = GraphQLObjectType("BadObject", {"badField": field})
        return GraphQLSchema(
            GraphQLObjectType("Query", {"f": GraphQLField(bad_object_type)}),
            types=[SomeObjectType],
        )

    @mark.parametrize("type_", output_types, ids=get_name)
    def accepts_an_output_type_as_an_object_field_type(type_):
        schema = _schema_with_object_field(type_)
        assert validate_schema(schema) == []

    def rejects_an_empty_object_field_type():
        # noinspection PyTypeChecker
        schema = _schema_with_object_field(None)  # type: ignore
        assert validate_schema(schema) == [
            {
                "message": "The type of BadObject.badField must be Output Type"
                " but got: None."
            }
        ]

    @mark.parametrize("type_", not_output_types, ids=get_name)
    def rejects_a_non_output_type_as_an_object_field_type(type_):
        schema = _schema_with_object_field(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of BadObject.badField must be Output Type"
                f" but got: {type_}."
            }
        ]

    @mark.parametrize("type_", not_graphql_types, ids=get_name)
    def rejects_a_non_type_value_as_an_object_field_type(type_):
        schema = _schema_with_object_field(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of BadObject.badField must be Output Type"
                f" but got: {inspect(type_)}.",
            },
            {"message": f"Expected GraphQL named type but got: {inspect(type_)}."},
        ]

    def rejects_with_relevant_locations_for_a_non_output_type():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  field: [SomeInputObject]
                }

                input SomeInputObject {
                  field: String
                }
                """
            )
        assert str(exc_info.value) == (
            "Query fields cannot be resolved. Field type must be an output type."
        )
        # therefore we need to monkey-patch a valid schema
        schema = build_schema(
            """
            type Query {
              field: [String]
            }

            input SomeInputObject {
              field: String
            }
            """
        )
        some_input_obj = schema.get_type("SomeInputObject")
        schema.query_type.fields["field"].type.of_type = some_input_obj  # type: ignore
        assert validate_schema(schema) == [
            {
                "message": "The type of Query.field must be Output Type"
                " but got: [SomeInputObject].",
                "locations": [(3, 22)],
            }
        ]


def describe_type_system_objects_can_only_implement_unique_interfaces():
    def rejects_an_object_implementing_a_non_type_value():
        query_type = GraphQLObjectType(
            "BadObject",
            {"f": GraphQLField(GraphQLString)},
        )
        # noinspection PyTypeChecker
        query_type.interfaces = (None,)
        schema = GraphQLSchema(query_type)

        assert validate_schema(schema) == [
            {
                "message": "Type BadObject must only implement Interface types,"
                " it cannot implement None."
            }
        ]

    def rejects_an_object_implementing_a_non_interface_type():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  test: BadObject
                }

                input SomeInputObject {
                  field: String
                }

                type BadObject implements SomeInputObject {
                  field: String
                }
                """
            )
        assert str(exc_info.value) == (
            "BadObject interfaces must be specified"
            " as a collection of GraphQLInterfaceType instances."
        )

    def rejects_an_object_implementing_the_same_interface_twice():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface & AnotherInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Type AnotherObject can only implement"
                " AnotherInterface once.",
                "locations": [(10, 43), (10, 62)],
            }
        ]

    def rejects_an_object_implementing_same_interface_twice_due_to_extension():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        extended_schema = extend_schema(
            schema, parse("extend type AnotherObject implements AnotherInterface")
        )
        assert validate_schema(extended_schema) == [
            {
                "message": "Type AnotherObject can only implement"
                " AnotherInterface once.",
                "locations": [(10, 43), (1, 38)],
            }
        ]


def describe_type_system_interface_extensions_should_be_valid():
    def rejects_object_implementing_extended_interface_due_to_missing_field():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        extended_schema = extend_schema(
            schema,
            parse(
                """
                extend interface AnotherInterface {
                  newField: String
                }

                extend type AnotherObject {
                  differentNewField: String
                }
                """
            ),
        )
        assert validate_schema(extended_schema) == [
            {
                "message": "Interface field AnotherInterface.newField expected"
                " but AnotherObject does not provide it.",
                "locations": [(3, 19), (10, 13), (6, 17)],
            }
        ]

    def rejects_object_implementing_extended_interface_due_to_missing_args():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        extended_schema = extend_schema(
            schema,
            parse(
                """
                extend interface AnotherInterface {
                  newField(test: Boolean): String
                }

                extend type AnotherObject {
                  newField: String
                }
                """
            ),
        )
        assert validate_schema(extended_schema) == [
            {
                "message": "Interface field argument"
                " AnotherInterface.newField(test:) expected"
                " but AnotherObject.newField does not provide it.",
                "locations": [(3, 28), (7, 19)],
            }
        ]

    def rejects_object_implementing_extended_interface_due_to_type_mismatch():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        extended_schema = extend_schema(
            schema,
            parse(
                """
                extend interface AnotherInterface {
                  newInterfaceField: NewInterface
                }

                interface NewInterface {
                  newField: String
                }

                interface MismatchingInterface {
                  newField: String
                }

                extend type AnotherObject {
                  newInterfaceField: MismatchingInterface
                }

                # Required to prevent unused interface errors
                type DummyObject implements NewInterface & MismatchingInterface {
                  newField: String
                }
                """
            ),
        )
        assert validate_schema(extended_schema) == [
            {
                "message": "Interface field AnotherInterface.newInterfaceField"
                " expects type NewInterface"
                " but AnotherObject.newInterfaceField"
                " is type MismatchingInterface.",
                "locations": [(3, 38), (15, 38)],
            }
        ]


def describe_type_system_interface_fields_must_have_output_types():
    def _schema_with_interface_field(type_: GraphQLOutputType) -> GraphQLSchema:
        if is_output_type(type_):
            field = GraphQLField(type_)
        else:
            # invalid field cannot be built with Python directly
            with raises(TypeError) as exc_info:
                GraphQLField(type_)
            assert str(exc_info.value) == "Field type must be an output type."
            # therefore we need to monkey-patch a valid field
            field = GraphQLField(GraphQLString)
            field.type = type_
        fields = {"badField": field}

        bad_interface_type = GraphQLInterfaceType("BadInterface", fields)
        bad_implementing_type = GraphQLObjectType(
            "BadImplementing",
            fields,
            interfaces=[bad_interface_type],
        )
        return GraphQLSchema(
            GraphQLObjectType("Query", {"f": GraphQLField(bad_interface_type)}),
            types=[bad_implementing_type, SomeObjectType],
        )

    @mark.parametrize("type_", output_types, ids=get_name)
    def accepts_an_output_type_as_an_interface_field_type(type_):
        schema = _schema_with_interface_field(type_)
        assert validate_schema(schema) == []

    def rejects_an_empty_interface_field_type():
        # noinspection PyTypeChecker
        schema = _schema_with_interface_field(None)  # type: ignore
        assert validate_schema(schema) == [
            {
                "message": "The type of BadImplementing.badField must be Output Type"
                " but got: None.",
            },
            {
                "message": "The type of BadInterface.badField must be Output Type"
                " but got: None.",
            },
        ]

    @mark.parametrize("type_", not_output_types, ids=get_name)
    def rejects_a_non_output_type_as_an_interface_field_type(type_):
        schema = _schema_with_interface_field(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of BadImplementing.badField must be Output Type"
                f" but got: {type_}.",
            },
            {
                "message": "The type of BadInterface.badField must be Output Type"
                f" but got: {type_}.",
            },
        ]

    @mark.parametrize("type_", not_graphql_types, ids=get_name)
    def rejects_a_non_type_value_as_an_interface_field_type(type_):
        schema = _schema_with_interface_field(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of BadImplementing.badField must be Output Type"
                f" but got: {inspect(type_)}.",
            },
            {
                "message": "The type of BadInterface.badField must be Output Type"
                f" but got: {inspect(type_)}.",
            },
            {"message": f"Expected GraphQL named type but got: {inspect(type_)}."},
        ]

    def rejects_a_non_output_type_as_an_interface_field_with_locations():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  test: SomeInterface
                }

                interface SomeInterface {
                  field: SomeInputObject
                }

                input SomeInputObject {
                  foo: String
                }

                type SomeObject implements SomeInterface {
                  field: SomeInputObject
                }
                """
            )
        assert str(exc_info.value) == (
            "SomeInterface fields cannot be resolved."
            " Field type must be an output type."
        )
        # therefore we need to monkey-patch a valid schema
        schema = build_schema(
            """
            type Query {
              test: SomeInterface
            }

            interface SomeInterface {
              field: String
            }

            input SomeInputObject {
              foo: String
            }

            type SomeObject implements SomeInterface {
              field: String
            }
            """
        )
        # therefore we need to monkey-patch a valid schema
        some_input_obj = schema.get_type("SomeInputObject")
        some_interface: Any = schema.get_type("SomeInterface")
        some_interface.fields["field"].type = some_input_obj
        some_object: Any = schema.get_type("SomeObject")
        some_object.fields["field"].type = some_input_obj
        assert validate_schema(schema) == [
            {
                "message": "The type of SomeInterface.field must be Output Type"
                " but got: SomeInputObject.",
                "locations": [(7, 22)],
            },
            {
                "message": "The type of SomeObject.field must be Output Type"
                " but got: SomeInputObject.",
                "locations": [(15, 22)],
            },
        ]

    def accepts_an_interface_not_implemented_by_at_least_one_object():
        schema = build_schema(
            """
            type Query {
              test: SomeInterface
            }

            interface SomeInterface {
              foo: String
            }
            """
        )
        assert validate_schema(schema) == []


def describe_type_system_arguments_must_have_input_types():
    def _schema_with_arg(type_: GraphQLInputType) -> GraphQLSchema:
        if is_input_type(type_):
            argument = GraphQLArgument(type_)
        else:
            # invalid argument cannot be built with Python directly
            with raises(TypeError) as exc_info:
                GraphQLArgument(type_)
            assert str(exc_info.value) == "Argument type must be a GraphQL input type."
            # therefore we need to monkey-patch a valid argument
            argument = GraphQLArgument(GraphQLString)
            argument.type = type_
        args = {"badArg": argument}
        bad_object_type = GraphQLObjectType(
            "BadObject",
            {"badField": GraphQLField(GraphQLString, args)},
        )
        return GraphQLSchema(
            GraphQLObjectType("Query", {"f": GraphQLField(bad_object_type)}),
            directives=[
                GraphQLDirective(
                    "BadDirective",
                    [DirectiveLocation.QUERY],
                    args,
                )
            ],
        )

    @mark.parametrize("type_", input_types, ids=get_name)
    def accepts_an_input_type_as_a_field_arg_type(type_):
        schema = _schema_with_arg(type_)
        assert validate_schema(schema) == []

    def rejects_an_empty_field_arg_type():
        # noinspection PyTypeChecker
        schema = _schema_with_arg(None)  # type: ignore
        assert validate_schema(schema) == [
            {
                "message": "The type of @BadDirective(badArg:) must be Input Type"
                " but got: None."
            },
            {
                "message": "The type of BadObject.badField(badArg:) must be Input Type"
                " but got: None."
            },
        ]

    @mark.parametrize("type_", not_input_types, ids=get_name)
    def rejects_a_non_input_type_as_a_field_arg_type(type_):
        schema = _schema_with_arg(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of @BadDirective(badArg:) must be Input Type"
                f" but got: {type_}."
            },
            {
                "message": "The type of BadObject.badField(badArg:) must be Input Type"
                f" but got: {type_}."
            },
        ]

    @mark.parametrize("type_", not_graphql_types, ids=get_name)
    def rejects_a_non_type_value_as_a_field_arg_type(type_):
        schema = _schema_with_arg(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of @BadDirective(badArg:) must be Input Type"
                f" but got: {inspect(type_)}."
            },
            {
                "message": "The type of BadObject.badField(badArg:) must be Input Type"
                f" but got: {inspect(type_)}."
            },
            {"message": f"Expected GraphQL named type but got: {inspect(type_)}."},
        ]

    def rejects_a_required_argument_that_is_deprecated():
        schema = build_schema(
            """
            directive @BadDirective(
              badArg: String! @deprecated
              optionalArg: String @deprecated
              anotherOptionalArg: String! = "" @deprecated
            ) on FIELD

            type Query {
              test(
                badArg: String! @deprecated
                optionalArg: String @deprecated
                anotherOptionalArg: String! = "" @deprecated
              ): String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Required argument @BadDirective(badArg:)"
                " cannot be deprecated.",
                "locations": [(3, 31), (3, 23)],
            },
            {
                "message": "Required argument Query.test(badArg:)"
                " cannot be deprecated.",
                "locations": [(10, 33), (10, 25)],
            },
        ]

    def rejects_a_non_input_type_as_a_field_arg_with_locations():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  test(arg: SomeObject): String
                }

                type SomeObject {
                  foo: String
                }
                """
            )
        assert str(exc_info.value) == (
            "Query fields cannot be resolved."
            " Argument type must be a GraphQL input type."
        )
        # therefore we need to monkey-patch a valid schema
        schema = build_schema(
            """
            type Query {
              test(arg: String): String
            }

            type SomeObject {
              foo: String
            }
            """
        )
        some_object = schema.get_type("SomeObject")
        schema.query_type.fields["test"].args["arg"].type = some_object  # type: ignore
        assert validate_schema(schema) == [
            {
                "message": "The type of Query.test(arg:) must be Input Type"
                " but got: SomeObject.",
                "locations": [(3, 25)],
            },
        ]


def describe_type_system_input_object_fields_must_have_input_types():
    def _schema_with_input_field(type_: GraphQLInputType) -> GraphQLSchema:
        if is_input_type(type_):
            input_field = GraphQLInputField(type_)
        else:
            # invalid input field cannot be built with Python directly
            with raises(TypeError) as exc_info:
                GraphQLInputField(type_)
            assert str(exc_info.value) == (
                "Input field type must be a GraphQL input type."
            )
            # therefore we need to monkey-patch a valid input field
            input_field = GraphQLInputField(GraphQLString)
            input_field.type = type_
        bad_input_object_type = GraphQLInputObjectType(
            "BadInputObject", {"badField": input_field}
        )
        return GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "f": GraphQLField(
                        GraphQLString,
                        args={"badArg": GraphQLArgument(bad_input_object_type)},
                    )
                },
            )
        )

    @mark.parametrize("type_", input_types, ids=get_name)
    def accepts_an_input_type_as_an_input_field_type(type_):
        schema = _schema_with_input_field(type_)
        assert validate_schema(schema) == []

    def rejects_an_empty_input_field_type():
        # noinspection PyTypeChecker
        schema = _schema_with_input_field(None)  # type: ignore
        assert validate_schema(schema) == [
            {
                "message": "The type of BadInputObject.badField must be Input Type"
                " but got: None."
            }
        ]

    @mark.parametrize("type_", not_input_types, ids=get_name)
    def rejects_a_non_input_type_as_an_input_field_type(type_):
        schema = _schema_with_input_field(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of BadInputObject.badField must be Input Type"
                f" but got: {type_}."
            }
        ]

    @mark.parametrize("type_", not_graphql_types, ids=get_name)
    def rejects_a_non_type_value_as_an_input_field_type(type_):
        schema = _schema_with_input_field(type_)
        assert validate_schema(schema) == [
            {
                "message": "The type of BadInputObject.badField must be Input Type"
                f" but got: {inspect(type_)}."
            },
            {"message": f"Expected GraphQL named type but got: {inspect(type_)}."},
        ]

    def rejects_with_relevant_locations_for_a_non_input_type():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  test(arg: SomeInputObject): String
                }

                input SomeInputObject {
                  foo: SomeObject
                }

                type SomeObject {
                  bar: String
                }
                """
            )
        assert str(exc_info.value) == (
            "SomeInputObject fields cannot be resolved."
            " Input field type must be a GraphQL input type."
        )
        # therefore we need to monkey-patch a valid schema
        schema = build_schema(
            """
            type Query {
              test(arg: SomeInputObject): String
            }

            input SomeInputObject {
              foo: String
            }

            type SomeObject {
              bar: String
            }
            """
        )
        some_object = schema.get_type("SomeObject")
        some_input_object: Any = schema.get_type("SomeInputObject")
        some_input_object.fields["foo"].type = some_object
        assert validate_schema(schema) == [
            {
                "message": "The type of SomeInputObject.foo must be Input Type"
                " but got: SomeObject.",
                "locations": [(7, 20)],
            }
        ]


def describe_objects_must_adhere_to_interfaces_they_implement():
    def accepts_an_object_which_implements_an_interface():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(input: String): String
            }
            """
        )
        assert validate_schema(schema) == []

    def accepts_an_object_which_implements_an_interface_and_with_more_fields():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(input: String): String
              anotherField: String
            }
            """
        )
        assert validate_schema(schema) == []

    def accepts_an_object_which_implements_an_interface_field_with_more_args():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(input: String, anotherInput: String): String
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_object_missing_an_interface_field():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              anotherField: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field expected but"
                " AnotherObject does not provide it.",
                "locations": [(7, 15), (10, 13)],
            }
        ]

    def rejects_an_object_with_an_incorrectly_typed_interface_field():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(input: String): Int
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field"
                " expects type String but"
                " AnotherObject.field is type Int.",
                "locations": [(7, 37), (11, 37)],
            }
        ]

    def rejects_an_object_with_a_differently_typed_interface_field():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            type A { foo: String }
            type B { foo: String }

            interface AnotherInterface {
              field: A
            }

            type AnotherObject implements AnotherInterface {
              field: B
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field"
                " expects type A but AnotherObject.field is type B.",
                "locations": [(10, 22), (14, 22)],
            }
        ]

    def accepts_an_object_with_a_subtyped_interface_field_interface():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: AnotherInterface
            }

            type AnotherObject implements AnotherInterface {
              field: AnotherObject
            }
            """
        )
        assert validate_schema(schema) == []

    def accepts_an_object_with_a_subtyped_interface_field_union():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            type SomeObject {
              field: String
            }

            union SomeUnionType = SomeObject

            interface AnotherInterface {
              field: SomeUnionType
            }

            type AnotherObject implements AnotherInterface {
              field: SomeObject
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_object_missing_an_interface_argument():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field argument"
                " AnotherInterface.field(input:) expected"
                " but AnotherObject.field does not provide it.",
                "locations": [(7, 21), (11, 15)],
            }
        ]

    def rejects_an_object_with_an_incorrectly_typed_interface_argument():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(input: Int): String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field argument"
                " AnotherInterface.field(input:) expects type String"
                " but AnotherObject.field(input:) is type Int.",
                "locations": [(7, 28), (11, 28)],
            }
        ]

    def rejects_an_object_with_an_incorrectly_typed_field_and_argument():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(input: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(input: Int): Int
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field expects"
                " type String but AnotherObject.field is type Int.",
                "locations": [(7, 37), (11, 34)],
            },
            {
                "message": "Interface field argument"
                " AnotherInterface.field(input:) expects type String"
                " but AnotherObject.field(input:) is type Int.",
                "locations": [(7, 28), (11, 28)],
            },
        ]

    def rejects_object_implementing_an_interface_field_with_additional_args():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field(baseArg: String): String
            }

            type AnotherObject implements AnotherInterface {
              field(
                baseArg: String,
                requiredArg: String!
                optionalArg1: String,
                optionalArg2: String = "",
              ): String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Object field AnotherObject.field includes required"
                " argument requiredArg that is missing from the"
                " Interface field AnotherInterface.field.",
                "locations": [(13, 17), (7, 15)],
            }
        ]

    def accepts_an_object_with_an_equivalently_wrapped_interface_field_type():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: [String]!
            }

            type AnotherObject implements AnotherInterface {
              field: [String]!
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_object_with_a_non_list_interface_field_list_type():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: [String]
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field expects type"
                " [String] but AnotherObject.field is type String.",
                "locations": [(7, 22), (11, 22)],
            }
        ]

    def rejects_an_object_with_a_list_interface_field_non_list_type():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface {
              field: [String]
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field expects type"
                " String but AnotherObject.field is type [String].",
                "locations": [(7, 22), (11, 22)],
            }
        ]

    def accepts_an_object_with_a_subset_non_null_interface_field_type():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String
            }

            type AnotherObject implements AnotherInterface {
              field: String!
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_object_with_a_superset_nullable_interface_field_type():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface AnotherInterface {
              field: String!
            }

            type AnotherObject implements AnotherInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field AnotherInterface.field expects type"
                " String! but AnotherObject.field is type String.",
                "locations": [(7, 22), (11, 22)],
            }
        ]

    def rejects_an_object_missing_a_transitive_interface():
        schema = build_schema(
            """
            type Query {
              test: AnotherObject
            }

            interface SuperInterface {
              field: String!
            }

            interface AnotherInterface implements SuperInterface {
              field: String!
            }

            type AnotherObject implements AnotherInterface {
              field: String!
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Type AnotherObject must implement SuperInterface"
                " because it is implemented by AnotherInterface.",
                "locations": [(10, 51), (14, 43)],
            }
        ]


def describe_interfaces_must_adhere_to_interface_they_implement():
    def accepts_an_interface_which_implements_an_interface():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(input: String): String
            }
            """
        )
        assert validate_schema(schema) == []

    def accepts_an_interface_which_implements_an_interface_along_with_more_fields():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(input: String): String
              anotherField: String
            }
            """
        )
        assert validate_schema(schema) == []

    def accepts_an_interface_which_implements_an_interface_with_additional_args():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(input: String, anotherInput: String): String
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_interface_missing_an_interface_field():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              anotherField: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field expected"
                " but ChildInterface does not provide it.",
                "locations": [(7, 15), (10, 13)],
            }
        ]

    def rejects_an_interface_with_an_incorrectly_typed_interface_field():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(input: String): Int
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field expects type String"
                " but ChildInterface.field is type Int.",
                "locations": [(7, 37), (11, 37)],
            }
        ]

    def rejects_an_interface_with_a_differently_typed_interface_field():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            type A { foo: String }
            type B { foo: String }

            interface ParentInterface {
              field: A
            }

            interface ChildInterface implements ParentInterface {
              field: B
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field expects type A"
                " but ChildInterface.field is type B.",
                "locations": [(10, 22), (14, 22)],
            }
        ]

    def accepts_an_interface_with_a_subtyped_interface_field_interface():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field: ParentInterface
            }

            interface ChildInterface implements ParentInterface {
              field: ChildInterface
            }
            """
        )
        assert validate_schema(schema) == []

    def accepts_an_interface_with_a_subtyped_interface_field_union():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            type SomeObject {
              field: String
            }

            union SomeUnionType = SomeObject

            interface ParentInterface {
              field: SomeUnionType
            }

            interface ChildInterface implements ParentInterface {
              field: SomeObject
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_interface_implementing_a_non_interface_type():
        # invalid schema cannot be built with Python
        with raises(TypeError) as exc_info:
            build_schema(
                """
                type Query {
                  field: String
                }

                input SomeInputObject {
                  field: String
                }

                interface BadInterface implements SomeInputObject {
                  field: String
                }
                """
            )
        assert str(exc_info.value) == (
            "BadInterface interfaces must be specified as a collection"
            " of GraphQLInterfaceType instances."
        )
        # therefore we construct the invalid schema manually
        some_input_obj = GraphQLInputObjectType(
            "SomeInputObject", {"field": GraphQLInputField(GraphQLString)}
        )
        bad_interface = GraphQLInterfaceType(
            "BadInterface", {"field": GraphQLField(GraphQLString)}
        )
        # noinspection PyTypeChecker
        bad_interface.interfaces = (some_input_obj,)
        schema = GraphQLSchema(
            GraphQLObjectType("Query", {"field": GraphQLField(GraphQLString)}),
            types=[bad_interface],
        )
        assert validate_schema(schema) == [
            {
                "message": "Type BadInterface must only implement Interface types,"
                " it cannot implement SomeInputObject.",
            }
        ]

    def rejects_an_interface_missing_an_interface_argument():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field argument ParentInterface.field(input:)"
                " expected but ChildInterface.field does not provide it.",
                "locations": [(7, 21), (11, 15)],
            }
        ]

    def rejects_an_interface_with_an_incorrectly_typed_interface_argument():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(input: Int): String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field argument ParentInterface.field(input:)"
                " expects type String but ChildInterface.field(input:) is type Int.",
                "locations": [(7, 28), (11, 28)],
            }
        ]

    def rejects_an_interface_with_both_an_incorrectly_typed_field_and_argument():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(input: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(input: Int): Int
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field expects type String"
                " but ChildInterface.field is type Int.",
                "locations": [(7, 37), (11, 34)],
            },
            {
                "message": "Interface field argument ParentInterface.field(input:)"
                " expects type String but ChildInterface.field(input:) is type Int.",
                "locations": [(7, 28), (11, 28)],
            },
        ]

    def rejects_an_interface_implementing_an_interface_field_with_additional_args():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field(baseArg: String): String
            }

            interface ChildInterface implements ParentInterface {
              field(
                baseArg: String,
                requiredArg: String!
                optionalArg1: String,
                optionalArg2: String = "",
              ): String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Object field ChildInterface.field includes"
                " required argument requiredArg that is missing"
                " from the Interface field ParentInterface.field.",
                "locations": [(13, 17), (7, 15)],
            }
        ]

    def accepts_an_interface_with_an_equivalently_wrapped_interface_field_type():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field: [String]!
            }

            interface ChildInterface implements ParentInterface {
              field: [String]!
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_interface_with_a_non_list_interface_field_list_type():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field: [String]
            }

            interface ChildInterface implements ParentInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field"
                " expects type [String] but ChildInterface.field is type String.",
                "locations": [(7, 22), (11, 22)],
            }
        ]

    def rejects_an_interface_with_a_list_interface_field_non_list_type():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field: String
            }

            interface ChildInterface implements ParentInterface {
              field: [String]
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field expects type String"
                " but ChildInterface.field is type [String].",
                "locations": [(7, 22), (11, 22)],
            }
        ]

    def accepts_an_interface_with_a_subset_non_null_interface_field_type():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field: String
            }

            interface ChildInterface implements ParentInterface {
              field: String!
            }
            """
        )
        assert validate_schema(schema) == []

    def rejects_an_interface_with_a_superset_nullable_interface_field_type():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface ParentInterface {
              field: String!
            }

            interface ChildInterface implements ParentInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Interface field ParentInterface.field expects type String!"
                " but ChildInterface.field is type String.",
                "locations": [(7, 22), (11, 22)],
            }
        ]

    def rejects_an_object_missing_a_transitive_interface():
        schema = build_schema(
            """
            type Query {
              test: ChildInterface
            }

            interface SuperInterface {
              field: String!
            }

            interface ParentInterface implements SuperInterface {
              field: String!
            }

            interface ChildInterface implements ParentInterface {
              field: String!
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Type ChildInterface must implement SuperInterface"
                " because it is implemented by ParentInterface.",
                "locations": [(10, 50), (14, 49)],
            }
        ]

    def rejects_a_self_reference_interface():
        schema = build_schema(
            """
            type Query {
            test: FooInterface
            }

            interface FooInterface implements FooInterface {
            field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Type FooInterface cannot implement itself"
                " because it would create a circular reference.",
                "locations": [(6, 47)],
            }
        ]

    def rejects_a_circular_interface_implementation():
        schema = build_schema(
            """
            type Query {
              test: FooInterface
            }

            interface FooInterface implements BarInterface {
              field: String
            }

            interface BarInterface implements FooInterface {
              field: String
            }
            """
        )
        assert validate_schema(schema) == [
            {
                "message": "Type FooInterface cannot implement BarInterface"
                " because it would create a circular reference.",
                "locations": [(10, 47), (6, 47)],
            },
            {
                "message": "Type BarInterface cannot implement FooInterface"
                " because it would create a circular reference.",
                "locations": [(6, 47), (10, 47)],
            },
        ]


def describe_assert_valid_schema():
    def does_not_throw_on_valid_schemas():
        schema = build_schema(
            (
                """
             type Query {
               foo: String
             }
            """
            )
        )
        assert_valid_schema(schema)

    def combines_multiple_errors():
        schema = build_schema("type SomeType")
        with raises(TypeError) as exc_info:
            assert_valid_schema(schema)
        assert (
            str(exc_info.value)
            == dedent(
                """
            Query root type must be provided.

            Type SomeType must define one or more fields.
            """
            ).rstrip()
        )



from functools import partial

from graphql.validation.rules.unique_argument_definition_names import (
    UniqueArgumentDefinitionNamesRule,
)

from .harness import assert_sdl_validation_errors

assert_sdl_errors = partial(
    assert_sdl_validation_errors, UniqueArgumentDefinitionNamesRule
)

assert_sdl_valid = partial(assert_sdl_errors, errors=[])


def describe_validate_unique_argument_definition_names():
    def no_args():
        assert_sdl_valid(
            """
            type SomeObject {
              someField: String
            }

            interface SomeInterface {
              someField: String
            }

            directive @someDirective on QUERY
            """
        )

    def one_argument():
        assert_sdl_valid(
            """
            type SomeObject {
              someField(foo: String): String
            }

            interface SomeInterface {
              someField(foo: String): String
            }

            extend type SomeObject {
              anotherField(foo: String): String
            }

            extend interface SomeInterface {
              anotherField(foo: String): String
            }

            directive @someDirective(foo: String) on QUERY
            """
        )

    def multiple_arguments():
        assert_sdl_valid(
            """
            type SomeObject {
              someField(
                foo: String
                bar: String
              ): String
            }

            interface SomeInterface {
              someField(
                foo: String
                bar: String
              ): String
            }

            extend type SomeObject {
              anotherField(
                foo: String
                bar: String
              ): String
            }

            extend interface SomeInterface {
              anotherField(
                foo: String
                bar: String
              ): String
            }

            directive @someDirective(
              foo: String
              bar: String
            ) on QUERY
            """
        )

    def duplicating_arguments():
        assert_sdl_errors(
            """
            type SomeObject {
              someField(
                foo: String
                bar: String
                foo: String
              ): String
            }

            interface SomeInterface {
              someField(
                foo: String
                bar: String
                foo: String
              ): String
            }

            extend type SomeObject {
              anotherField(
                foo: String
                bar: String
                bar: String
              ): String
            }

            extend interface SomeInterface {
              anotherField(
                bar: String
                foo: String
                foo: String
              ): String
            }

            directive @someDirective(
              foo: String
              bar: String
              foo: String
            ) on QUERY
            """,
            [
                {
                    "message": "Argument 'SomeObject.someField(foo:)'"
                    " can only be defined once.",
                    "locations": [(4, 17), (6, 17)],
                },
                {
                    "message": "Argument 'SomeInterface.someField(foo:)'"
                    " can only be defined once.",
                    "locations": [(12, 17), (14, 17)],
                },
                {
                    "message": "Argument 'SomeObject.anotherField(bar:)'"
                    " can only be defined once.",
                    "locations": [(21, 17), (22, 17)],
                },
                {
                    "message": "Argument 'SomeInterface.anotherField(foo:)'"
                    " can only be defined once.",
                    "locations": [(29, 17), (30, 17)],
                },
                {
                    "message": "Argument '@someDirective(foo:)'"
                    " can only be defined once.",
                    "locations": [(35, 15), (37, 15)],
                },
            ],
        )



from functools import partial

from graphql.utilities import build_schema
from graphql.validation import PossibleFragmentSpreadsRule

from .harness import assert_validation_errors

test_schema = build_schema(
    """
    interface Being {
      name: String
    }

    interface Pet implements Being {
      name: String
    }

    type Dog implements Being & Pet {
      name: String
      barkVolume: Int
    }

    type Cat implements Being & Pet {
      name: String
      meowVolume: Int
    }

    union CatOrDog = Cat | Dog

    interface Intelligent {
      iq: Int
    }

    type Human implements Being & Intelligent {
      name: String
      pets: [Pet]
      iq: Int
    }

    type Alien implements Being & Intelligent {
      name: String
      iq: Int
    }

    union DogOrHuman = Dog | Human

    union HumanOrAlien = Human | Alien

    type Query {
      catOrDog: CatOrDog
      dogOrHuman: DogOrHuman
      humanOrAlien: HumanOrAlien
    }
    """
)


assert_errors = partial(
    assert_validation_errors, PossibleFragmentSpreadsRule, schema=test_schema
)

assert_valid = partial(assert_errors, errors=[])


def describe_validate_possible_fragment_spreads():
    def of_the_same_object():
        assert_valid(
            """
            fragment objectWithinObject on Dog { ...dogFragment }
            fragment dogFragment on Dog { barkVolume }
            """
        )

    def of_the_same_object_inline_fragment():
        assert_valid(
            """
            fragment objectWithinObjectAnon on Dog { ... on Dog { barkVolume } }
            """
        )

    def object_into_implemented_interface():
        assert_valid(
            """
            fragment objectWithinInterface on Pet { ...dogFragment }
            fragment dogFragment on Dog { barkVolume }
            """
        )

    def object_into_containing_union():
        assert_valid(
            """
            fragment objectWithinUnion on CatOrDog { ...dogFragment }
            fragment dogFragment on Dog { barkVolume }
            """
        )

    def union_into_contained_object():
        assert_valid(
            """
            fragment unionWithinObject on Dog { ...catOrDogFragment }
            fragment catOrDogFragment on CatOrDog { __typename }
            """
        )

    def union_into_overlapping_interface():
        assert_valid(
            """
            fragment unionWithinInterface on Pet { ...catOrDogFragment }
            fragment catOrDogFragment on CatOrDog { __typename }
            """
        )

    def union_into_overlapping_union():
        assert_valid(
            """
            fragment unionWithinUnion on DogOrHuman { ...catOrDogFragment }
            fragment catOrDogFragment on CatOrDog { __typename }
            """
        )

    def interface_into_implemented_object():
        assert_valid(
            """
            fragment interfaceWithinObject on Dog { ...petFragment }
            fragment petFragment on Pet { name }
            """
        )

    def interface_into_overlapping_interface():
        assert_valid(
            """
            fragment interfaceWithinInterface on Pet { ...beingFragment }
            fragment beingFragment on Being { name }
            """
        )

    def interface_into_overlapping_interface_in_inline_fragment():
        assert_valid(
            """
            fragment interfaceWithinInterface on Pet { ... on Being { name } }
            """
        )

    def interface_into_overlapping_union():
        assert_valid(
            """
            fragment interfaceWithinUnion on CatOrDog { ...petFragment }
            fragment petFragment on Pet { name }
            """
        )

    def ignores_incorrect_type_caught_by_fragments_on_composite_types():
        assert_valid(
            """
            fragment petFragment on Pet { ...badInADifferentWay }
            fragment badInADifferentWay on String { name }
            """
        )

    def ignores_unknown_fragments_caught_by_known_fragment_names():
        assert_valid(
            """
            fragment petFragment on Pet { ...UnknownFragment }
            """
        )

    def different_object_into_object():
        assert_errors(
            """
            fragment invalidObjectWithinObject on Cat { ...dogFragment }
            fragment dogFragment on Dog { barkVolume }
            """,
            [
                {
                    "message": "Fragment 'dogFragment' cannot be spread here"
                    " as objects of type 'Cat' can never be of type 'Dog'.",
                    "locations": [(2, 57)],
                },
            ],
        )

    def different_object_into_object_in_inline_fragment():
        assert_errors(
            """
            fragment invalidObjectWithinObjectAnon on Cat {
              ... on Dog { barkVolume }
            }
            """,
            [
                {
                    "message": "Fragment cannot be spread here"
                    " as objects of type 'Cat' can never be of type 'Dog'.",
                    "locations": [(3, 15)],
                },
            ],
        )

    def object_into_not_implementing_interface():
        assert_errors(
            """
            fragment invalidObjectWithinInterface on Pet { ...humanFragment }
            fragment humanFragment on Human { pets { name } }
            """,
            [
                {
                    "message": "Fragment 'humanFragment' cannot be spread here"
                    " as objects of type 'Pet' can never be of type 'Human'.",
                    "locations": [(2, 60)],
                },
            ],
        )

    def object_into_not_containing_union():
        assert_errors(
            """
            fragment invalidObjectWithinUnion on CatOrDog { ...humanFragment }
            fragment humanFragment on Human { pets { name } }
            """,
            [
                {
                    "message": "Fragment 'humanFragment' cannot be spread here"
                    " as objects of type 'CatOrDog' can never be of type 'Human'.",
                    "locations": [(2, 61)],
                },
            ],
        )

    def union_into_not_contained_object():
        assert_errors(
            """
            fragment invalidUnionWithinObject on Human { ...catOrDogFragment }
            fragment catOrDogFragment on CatOrDog { __typename }
            """,
            [
                {
                    "message": "Fragment 'catOrDogFragment' cannot be spread here"
                    " as objects of type 'Human' can never be of type 'CatOrDog'.",
                    "locations": [(2, 58)],
                },
            ],
        )

    def union_into_non_overlapping_interface():
        assert_errors(
            """
            fragment invalidUnionWithinInterface on Pet { ...humanOrAlienFragment }
            fragment humanOrAlienFragment on HumanOrAlien { __typename }
            """,
            [
                {
                    "message": "Fragment 'humanOrAlienFragment' cannot be spread here"
                    " as objects of type 'Pet' can never be of type 'HumanOrAlien'.",
                    "locations": [(2, 59)],
                },
            ],
        )

    def union_into_non_overlapping_union():
        assert_errors(
            """
            fragment invalidUnionWithinUnion on CatOrDog { ...humanOrAlienFragment }
            fragment humanOrAlienFragment on HumanOrAlien { __typename }
            """,
            [
                {
                    "message": "Fragment 'humanOrAlienFragment'"
                    " cannot be spread here as objects of type 'CatOrDog'"
                    " can never be of type 'HumanOrAlien'.",
                    "locations": [(2, 60)],
                },
            ],
        )

    def interface_into_non_implementing_object():
        assert_errors(
            """
            fragment invalidInterfaceWithinObject on Cat { ...intelligentFragment }
            fragment intelligentFragment on Intelligent { iq }
            """,
            [
                {
                    "message": "Fragment 'intelligentFragment' cannot be spread here"
                    " as objects of type 'Cat' can never be of type 'Intelligent'.",
                    "locations": [(2, 60)],
                },
            ],
        )

    def interface_into_non_overlapping_interface():
        assert_errors(
            """
            fragment invalidInterfaceWithinInterface on Pet {
              ...intelligentFragment
            }
            fragment intelligentFragment on Intelligent { iq }
            """,
            [
                {
                    "message": "Fragment 'intelligentFragment' cannot be spread here"
                    " as objects of type 'Pet' can never be of type 'Intelligent'.",
                    "locations": [(3, 15)],
                },
            ],
        )

    def interface_into_non_overlapping_interface_in_inline_fragment():
        assert_errors(
            """
            fragment invalidInterfaceWithinInterfaceAnon on Pet {
              ...on Intelligent { iq }
            }
            """,
            [
                {
                    "message": "Fragment cannot be spread here as objects"
                    " of type 'Pet' can never be of type 'Intelligent'.",
                    "locations": [(3, 15)],
                },
            ],
        )

    def interface_into_non_overlapping_union():
        assert_errors(
            """
            fragment invalidInterfaceWithinUnion on HumanOrAlien { ...petFragment }
            fragment petFragment on Pet { name }
            """,
            [
                {
                    "message": "Fragment 'petFragment' cannot be spread here"
                    " as objects of type 'HumanOrAlien' can never be of type 'Pet'.",
                    "locations": [(2, 68)],
                },
            ],
        )



from typing import cast, Any, TypeVar

from .undefined import Undefined

__all__ = ["identity_func"]


T = TypeVar("T")


def identity_func(x: T = cast(Any, Undefined), *_args: Any) -> T:
    """Return the first received argument."""
    return x



"""GraphQL Language

The :mod:`graphql.language` package is responsible for parsing and operating on the
GraphQL language.
"""

from .source import Source

from .location import get_location, SourceLocation, FormattedSourceLocation

from .print_location import print_location, print_source_location

from .token_kind import TokenKind

from .lexer import Lexer

from .parser import parse, parse_type, parse_value, parse_const_value

from .printer import print_ast

from .visitor import (
    visit,
    Visitor,
    ParallelVisitor,
    VisitorAction,
    VisitorKeyMap,
    BREAK,
    SKIP,
    REMOVE,
    IDLE,
)

from .ast import (
    Location,
    Token,
    Node,
    # Each kind of AST node
    NameNode,
    DocumentNode,
    DefinitionNode,
    ExecutableDefinitionNode,
    OperationDefinitionNode,
    OperationType,
    VariableDefinitionNode,
    VariableNode,
    SelectionSetNode,
    SelectionNode,
    FieldNode,
    ArgumentNode,
    ConstArgumentNode,
    FragmentSpreadNode,
    InlineFragmentNode,
    FragmentDefinitionNode,
    ValueNode,
    ConstValueNode,
    IntValueNode,
    FloatValueNode,
    StringValueNode,
    BooleanValueNode,
    NullValueNode,
    EnumValueNode,
    ListValueNode,
    ConstListValueNode,
    ObjectValueNode,
    ConstObjectValueNode,
    ObjectFieldNode,
    ConstObjectFieldNode,
    DirectiveNode,
    ConstDirectiveNode,
    TypeNode,
    NamedTypeNode,
    ListTypeNode,
    NonNullTypeNode,
    TypeSystemDefinitionNode,
    SchemaDefinitionNode,
    OperationTypeDefinitionNode,
    TypeDefinitionNode,
    ScalarTypeDefinitionNode,
    ObjectTypeDefinitionNode,
    FieldDefinitionNode,
    InputValueDefinitionNode,
    InterfaceTypeDefinitionNode,
    UnionTypeDefinitionNode,
    EnumTypeDefinitionNode,
    EnumValueDefinitionNode,
    InputObjectTypeDefinitionNode,
    DirectiveDefinitionNode,
    TypeSystemExtensionNode,
    SchemaExtensionNode,
    TypeExtensionNode,
    ScalarTypeExtensionNode,
    ObjectTypeExtensionNode,
    InterfaceTypeExtensionNode,
    UnionTypeExtensionNode,
    EnumTypeExtensionNode,
    InputObjectTypeExtensionNode,
)
from .predicates import (
    is_definition_node,
    is_executable_definition_node,
    is_selection_node,
    is_value_node,
    is_const_value_node,
    is_type_node,
    is_type_system_definition_node,
    is_type_definition_node,
    is_type_system_extension_node,
    is_type_extension_node,
)
from .directive_locations import DirectiveLocation

__all__ = [
    "get_location",
    "SourceLocation",
    "FormattedSourceLocation",
    "print_location",
    "print_source_location",
    "TokenKind",
    "Lexer",
    "parse",
    "parse_value",
    "parse_const_value",
    "parse_type",
    "print_ast",
    "Source",
    "visit",
    "Visitor",
    "ParallelVisitor",
    "VisitorAction",
    "VisitorKeyMap",
    "BREAK",
    "SKIP",
    "REMOVE",
    "IDLE",
    "Location",
    "Token",
    "DirectiveLocation",
    "Node",
    "NameNode",
    "DocumentNode",
    "DefinitionNode",
    "ExecutableDefinitionNode",
    "OperationDefinitionNode",
    "OperationType",
    "VariableDefinitionNode",
    "VariableNode",
    "SelectionSetNode",
    "SelectionNode",
    "FieldNode",
    "ArgumentNode",
    "ConstArgumentNode",
    "FragmentSpreadNode",
    "InlineFragmentNode",
    "FragmentDefinitionNode",
    "ValueNode",
    "ConstValueNode",
    "IntValueNode",
    "FloatValueNode",
    "StringValueNode",
    "BooleanValueNode",
    "NullValueNode",
    "EnumValueNode",
    "ListValueNode",
    "ConstListValueNode",
    "ObjectValueNode",
    "ConstObjectValueNode",
    "ObjectFieldNode",
    "ConstObjectFieldNode",
    "DirectiveNode",
    "ConstDirectiveNode",
    "TypeNode",
    "NamedTypeNode",
    "ListTypeNode",
    "NonNullTypeNode",
    "TypeSystemDefinitionNode",
    "SchemaDefinitionNode",
    "OperationTypeDefinitionNode",
    "TypeDefinitionNode",
    "ScalarTypeDefinitionNode",
    "ObjectTypeDefinitionNode",
    "FieldDefinitionNode",
    "InputValueDefinitionNode",
    "InterfaceTypeDefinitionNode",
    "UnionTypeDefinitionNode",
    "EnumTypeDefinitionNode",
    "EnumValueDefinitionNode",
    "InputObjectTypeDefinitionNode",
    "DirectiveDefinitionNode",
    "TypeSystemExtensionNode",
    "SchemaExtensionNode",
    "TypeExtensionNode",
    "ScalarTypeExtensionNode",
    "ObjectTypeExtensionNode",
    "InterfaceTypeExtensionNode",
    "UnionTypeExtensionNode",
    "EnumTypeExtensionNode",
    "InputObjectTypeExtensionNode",
    "is_definition_node",
    "is_executable_definition_node",
    "is_selection_node",
    "is_value_node",
    "is_const_value_node",
    "is_type_node",
    "is_type_system_definition_node",
    "is_type_definition_node",
    "is_type_system_extension_node",
    "is_type_extension_node",
]



from pytest import raises

from graphql.error import GraphQLError
from graphql.execution import ExecutionResult


def describe_execution_result():

    data = {"foo": "Some data"}
    error = GraphQLError("Some error")
    errors = [error]
    extensions = {"bar": "Some extension"}

    def initializes_properly():
        res = ExecutionResult(data, errors)
        assert res.data is data
        assert res.errors is errors
        assert res.extensions is None
        res = ExecutionResult(data, errors, extensions)
        assert res.data is data
        assert res.errors is errors
        assert res.extensions is extensions

    def prints_a_representation():
        assert repr(ExecutionResult(data, errors)) == (
            "ExecutionResult(data={'foo': 'Some data'},"
            " errors=[GraphQLError('Some error')])"
        )
        assert repr(ExecutionResult(data, errors, extensions)) == (
            "ExecutionResult(data={'foo': 'Some data'},"
            " errors=[GraphQLError('Some error')],"
            " extensions={'bar': 'Some extension'})"
        )

    def formats_properly():
        res = ExecutionResult(data, None)
        assert res.formatted == {"data": data}
        res = ExecutionResult(data, errors)
        assert res.formatted == {
            "data": data,
            "errors": [{"message": "Some error"}],
        }
        res = ExecutionResult(data, None, extensions)
        assert res.formatted == {
            "data": data,
            "extensions": extensions,
        }
        res = ExecutionResult(data, errors, extensions)
        assert res.formatted == {
            "data": data,
            "errors": [{"message": "Some error"}],
            "extensions": extensions,
        }

    def compares_to_dict():
        res = ExecutionResult(data, errors)
        assert res == {"data": data, "errors": errors}
        assert res == {"data": data, "errors": errors, "extensions": None}
        assert res != {"data": data, "errors": None}
        assert res != {"data": None, "errors": errors}
        assert res != {"data": data, "errors": errors, "extensions": extensions}
        res = ExecutionResult(data, errors, extensions)
        assert res == {"data": data, "errors": errors}
        assert res == {"data": data, "errors": errors, "extensions": extensions}
        assert res != {"data": data, "errors": None}
        assert res != {"data": None, "errors": errors}
        assert res != {"data": data, "errors": errors, "extensions": None}

    def compares_to_tuple():
        res = ExecutionResult(data, errors)
        assert res == (data, errors)
        assert res == (data, errors, None)
        assert res != (data, None)
        assert res != (None, errors)
        assert res != (data, errors, extensions)
        res = ExecutionResult(data, errors, extensions)
        assert res == (data, errors)
        assert res == (data, errors, extensions)
        assert res != (data, None)
        assert res != (None, errors)
        assert res != (data, errors, None)

    def does_not_compare_to_list():
        res = ExecutionResult(data, errors)
        assert res != [data, errors]
        res = ExecutionResult(data, errors, extensions)
        assert res != [data, errors, extensions]

    def compares_to_another_execution_result():
        res1 = ExecutionResult(data, errors)
        res2 = ExecutionResult(data, errors)
        assert res1 == res2
        res2 = ExecutionResult({"foo": "other data"}, errors)
        assert res1 != res2
        res2 = ExecutionResult(data, [GraphQLError("Another error")])
        assert res1 != res2
        res2 = ExecutionResult(data, errors, extensions)
        assert res1 != res2
        res1 = ExecutionResult(data, errors, extensions)
        res2 = ExecutionResult(data, errors, extensions)
        assert res1 == res2
        res2 = ExecutionResult({"foo": "other data"}, errors, extensions)
        assert res1 != res2
        res2 = ExecutionResult(data, [GraphQLError("Another error")], extensions)
        assert res1 != res2
        res2 = ExecutionResult(data, errors, {"bar": "Another extension"})
        assert res1 != res2

    def unpacks_as_two_tuple():
        res = ExecutionResult(data, errors)
        res_data, res_errors = res  # type: ignore
        assert res_data == data  # type: ignore
        assert res_errors == errors  # type: ignore
        with raises(ValueError):
            res = ExecutionResult(data, errors, extensions)
            _res_data, _res_errors, _res_extensions = res  # type: ignore



from inspect import isawaitable
from typing import Any, NamedTuple, Optional

from pytest import mark

from graphql.execution import execute, execute_sync, ExecutionResult
from graphql.language import parse
from graphql.type import (
    GraphQLBoolean,
    GraphQLField,
    GraphQLInterfaceType,
    GraphQLList,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
    GraphQLUnionType,
)
from graphql.utilities import build_schema


def sync_and_async(spec):
    """Decorator for running a test synchronously and asynchronously."""
    return mark.asyncio(
        mark.parametrize("sync", (True, False), ids=("sync", "async"))(spec)
    )


def access_variants(spec):
    """Decorator for tests with dict and object access, including inheritance."""
    return mark.asyncio(
        mark.parametrize("access", ("dict", "object", "inheritance"))(spec)
    )


async def execute_query(
    sync: bool, schema: GraphQLSchema, query: str, root_value: Any = None
) -> ExecutionResult:
    """Execute the query against the given schema synchronously or asynchronously."""
    assert isinstance(sync, bool)
    assert isinstance(schema, GraphQLSchema)
    assert isinstance(query, str)
    document = parse(query)
    result = (execute_sync if sync else execute)(
        schema, document, root_value
    )  # type: ignore
    if not sync and isawaitable(result):
        result = await result
    assert isinstance(result, ExecutionResult)
    return result


def get_is_type_of(type_, sync=True):
    """Get a sync or async is_type_of function for the given type."""
    if sync:

        def is_type_of(obj, _info):
            return isinstance(obj, type_)

    else:

        async def is_type_of(obj, _info):
            return isinstance(obj, type_)

    return is_type_of


def get_type_error(sync=True):
    """Get a sync or async is_type_of or type resolver function that raises an error."""
    error = RuntimeError("We are testing this error")
    if sync:

        def type_error(*_args):
            raise error

    else:

        async def type_error(*_args):
            raise error

    return type_error


class Dog(NamedTuple):

    name: str
    woofs: bool


class Cat(NamedTuple):

    name: str
    meows: bool


def describe_execute_handles_synchronous_execution_of_abstract_types():
    @sync_and_async
    async def is_type_of_used_to_resolve_runtime_type_for_interface(sync):
        pet_type = GraphQLInterfaceType("Pet", {"name": GraphQLField(GraphQLString)})

        dog_type = GraphQLObjectType(
            "Dog",
            {
                "name": GraphQLField(GraphQLString),
                "woofs": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
            is_type_of=get_is_type_of(Dog, sync),
        )

        cat_type = GraphQLObjectType(
            "Cat",
            {
                "name": GraphQLField(GraphQLString),
                "meows": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
            is_type_of=get_is_type_of(Cat, sync),
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "pets": GraphQLField(
                        GraphQLList(pet_type),
                        resolve=lambda *_args: [
                            Dog("Odie", True),
                            Cat("Garfield", False),
                        ],
                    )
                },
            ),
            types=[cat_type, dog_type],
        )

        query = """
            {
              pets {
                name
                ... on Dog {
                  woofs
                }
                ... on Cat {
                  meows
                }
              }
            }
            """

        assert await execute_query(sync, schema, query) == (
            {
                "pets": [
                    {"name": "Odie", "woofs": True},
                    {"name": "Garfield", "meows": False},
                ]
            },
            None,
        )

    @sync_and_async
    async def is_type_of_can_throw(sync):
        pet_type = GraphQLInterfaceType("Pet", {"name": GraphQLField(GraphQLString)})

        dog_type = GraphQLObjectType(
            "Dog",
            {
                "name": GraphQLField(GraphQLString),
                "woofs": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
            is_type_of=get_type_error(sync),
        )

        cat_type = GraphQLObjectType(
            "Cat",
            {
                "name": GraphQLField(GraphQLString),
                "meows": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
            is_type_of=None,
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "pets": GraphQLField(
                        GraphQLList(pet_type),
                        resolve=lambda *_args: [
                            Dog("Odie", True),
                            Cat("Garfield", False),
                        ],
                    )
                },
            ),
            types=[dog_type, cat_type],
        )

        query = """
            {
              pets {
                name
                ... on Dog {
                  woofs
                }
                ... on Cat {
                  meows
                }
              }
            }
            """

        assert await execute_query(sync, schema, query) == (
            {"pets": [None, None]},
            [
                {
                    "message": "We are testing this error",
                    "locations": [(3, 15)],
                    "path": ["pets", 0],
                },
                {
                    "message": "We are testing this error",
                    "locations": [(3, 15)],
                    "path": ["pets", 1],
                },
            ],
        )

    @sync_and_async
    async def is_type_of_with_no_suitable_type(sync):
        # added in GraphQL-core to improve coverage
        pet_type = GraphQLInterfaceType("Pet", {"name": GraphQLField(GraphQLString)})

        dog_type = GraphQLObjectType(
            "Dog",
            {
                "name": GraphQLField(GraphQLString),
                "woofs": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
            is_type_of=get_is_type_of(Cat, sync),
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "pets": GraphQLField(
                        GraphQLList(pet_type),
                        resolve=lambda *_args: [Dog("Odie", True)],
                    )
                },
            ),
            types=[dog_type],
        )

        query = """
            {
              pets {
                name
                ... on Dog {
                  woofs
                }
              }
            }
            """

        message = (
            "Abstract type 'Pet' must resolve to an Object type at runtime"
            " for field 'Query.pets'."
            " Either the 'Pet' type should provide a 'resolve_type' function"
            " or each possible type should provide an 'is_type_of' function."
        )
        assert await execute_query(sync, schema, query) == (
            {"pets": [None]},
            [{"message": message, "locations": [(3, 15)], "path": ["pets", 0]}],
        )

    @sync_and_async
    async def is_type_of_used_to_resolve_runtime_type_for_union(sync):
        dog_type = GraphQLObjectType(
            "Dog",
            {
                "name": GraphQLField(GraphQLString),
                "woofs": GraphQLField(GraphQLBoolean),
            },
            is_type_of=get_is_type_of(Dog, sync),
        )

        cat_type = GraphQLObjectType(
            "Cat",
            {
                "name": GraphQLField(GraphQLString),
                "meows": GraphQLField(GraphQLBoolean),
            },
            is_type_of=get_is_type_of(Cat, sync),
        )

        pet_type = GraphQLUnionType("Pet", [cat_type, dog_type])

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "pets": GraphQLField(
                        GraphQLList(pet_type),
                        resolve=lambda *_args: [
                            Dog("Odie", True),
                            Cat("Garfield", False),
                        ],
                    )
                },
            )
        )

        query = """
            {
              pets {
                ... on Dog {
                  name
                  woofs
                }
                ... on Cat {
                  name
                  meows
                }
              }
            }
            """

        assert await execute_query(sync, schema, query) == (
            {
                "pets": [
                    {"name": "Odie", "woofs": True},
                    {"name": "Garfield", "meows": False},
                ]
            },
            None,
        )

    @sync_and_async
    async def resolve_type_can_throw(sync):
        pet_type = GraphQLInterfaceType(
            "Pet",
            {"name": GraphQLField(GraphQLString)},
            resolve_type=get_type_error(sync),
        )

        dog_type = GraphQLObjectType(
            "Dog",
            {
                "name": GraphQLField(GraphQLString),
                "woofs": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
        )

        cat_type = GraphQLObjectType(
            "Cat",
            {
                "name": GraphQLField(GraphQLString),
                "meows": GraphQLField(GraphQLBoolean),
            },
            interfaces=[pet_type],
        )

        schema = GraphQLSchema(
            GraphQLObjectType(
                "Query",
                {
                    "pets": GraphQLField(
                        GraphQLList(pet_type),
                        resolve=lambda *_args: [
                            Dog("Odie", True),
                            Cat("Garfield", False),
                        ],
                    )
                },
            ),
            types=[dog_type, cat_type],
        )

        query = """
            {
              pets {
                name
                ... on Dog {
                  woofs
                }
                ... on Cat {
                  meows
                }
              }
            }
            """

        assert await execute_query(sync, schema, query) == (
            {"pets": [None, None]},
            [
                {
                    "message": "We are testing this error",
                    "locations": [(3, 15)],
                    "path": ["pets", 0],
                },
                {
                    "message": "We are testing this error",
                    "locations": [(3, 15)],
                    "path": ["pets", 1],
                },
            ],
        )

    def describe_using_typename_on_source_object():

        expected = (
            {
                "pets": [
                    {"name": "Odie", "woofs": True},
                    {"name": "Garfield", "meows": False},
                ]
            },
            None,
        )

        # noinspection PyShadowingNames
        def _root_value(access: str) -> Any:
            if access == "dict":
                return {
                    "pets": [
                        {"__typename": "Dog", "name": "Odie", "woofs": True},
                        {"__typename": "Cat", "name": "Garfield", "meows": False},
                    ],
                }

            if access == "object":

                class DogObject:
                    __typename = "Dog"
                    name = "Odie"
                    woofs = True

                class CatObject:
                    __typename = "Cat"
                    name = "Garfield"
                    meows = False

                class RootValueAsObject:
                    pets = [DogObject(), CatObject()]

                return RootValueAsObject()

            if access == "inheritance":

                class Pet:
                    __typename = "Pet"
                    name: Optional[str] = None

                class DogPet(Pet):
                    __typename = "Dog"
                    woofs: Optional[bool] = None

                class Odie(DogPet):
                    name = "Odie"
                    woofs = True

                class CatPet(Pet):
                    __typename = "Cat"
                    meows: Optional[bool] = None

                class Tabby(CatPet):
                    pass

                class Garfield(Tabby):
                    name = "Garfield"
                    meows = False

                class RootValueWithInheritance:
                    pets = [Odie(), Garfield()]

                return RootValueWithInheritance()

            assert False, f"Unknown access variant: {access}"  # pragma: no cover

        def describe_union_type():

            schema = build_schema(
                """
                type Query {
                  pets: [Pet]
                }

                union Pet = Cat | Dog

                type Cat {
                  name: String
                  meows: Boolean
                }

                type Dog {
                  name: String
                  woofs: Boolean
                }
                """
            )

            query = """
                {
                  pets {
                    name
                    ... on Dog {
                      woofs
                    }
                    ... on Cat {
                      meows
                    }
                  }
                }
                """

            @sync_and_async
            @access_variants
            async def resolve(sync, access):
                root_value = _root_value(access)
                assert await execute_query(sync, schema, query, root_value) == expected

        def describe_interface_type():
            schema = build_schema(
                """
                type Query {
                  pets: [Pet]
                }

                interface Pet {
                  name: String
                  }

                type Cat implements Pet {
                  name: String
                  meows: Boolean
                }

                type Dog implements Pet {
                  name: String
                  woofs: Boolean
                }
                """
            )

            query = """
                {
                  pets {
                    name
                    ... on Dog {
                      woofs
                    }
                    ... on Cat {
                      meows
                    }
                  }
                }
                """

            @sync_and_async
            @access_variants
            async def resolve(sync, access):
                root_value = _root_value(access)
                assert await execute_query(sync, schema, query, root_value) == expected

    def resolve_type_on_interface_yields_useful_error():
        schema = build_schema(
            """
            type Query {
              pet: Pet
            }

            interface Pet {
              name: String
            }

            type Cat implements Pet {
              name: String
            }

            type Dog implements Pet {
              name: String
            }
            """
        )

        document = parse(
            """
            {
              pet {
                name
              }
            }
            """
        )

        def expect_error(for_type_name: Any, message: str) -> None:
            root_value = {"pet": {"__typename": for_type_name}}
            result = execute_sync(schema, document, root_value)
            expected = (
                {"pet": None},
                [{"message": message, "locations": [(3, 15)], "path": ["pet"]}],
            )
            assert result == expected

        expect_error(
            None,
            "Abstract type 'Pet' must resolve"
            " to an Object type at runtime for field 'Query.pet'."
            " Either the 'Pet' type should provide a 'resolve_type' function"
            " or each possible type should provide an 'is_type_of' function.",
        )

        expect_error(
            "Human",
            "Abstract type 'Pet' was resolved to a type 'Human'"
            " that does not exist inside the schema.",
        )

        expect_error(
            "String", "Abstract type 'Pet' was resolved to a non-object type 'String'."
        )

        expect_error(
            "__Schema",
            "Runtime Object type '__Schema' is not a possible type for 'Pet'.",
        )

        # workaround since we can't inject resolve_type into SDL
        schema.get_type("Pet").resolve_type = lambda *_args: []  # type: ignore

        expect_error(
            None,
            "Abstract type 'Pet' must resolve"
            " to an Object type at runtime for field 'Query.pet'"
            " with value {'__typename': None}, received '[]'.",
        )



from graphql.execution import execute_sync
from graphql.language import parse
from graphql.type import (
    GraphQLArgument,
    GraphQLBoolean,
    GraphQLField,
    GraphQLID,
    GraphQLInt,
    GraphQLList,
    GraphQLNonNull,
    GraphQLObjectType,
    GraphQLSchema,
    GraphQLString,
)


def describe_execute_handles_execution_with_a_complex_schema():
    def executes_using_a_schema():
        class Article:

            # noinspection PyShadowingBuiltins
            def __init__(self, id: int):
                self.id = id
                self.isPublished = True
                self.author = JohnSmith()
                self.title = f"My Article {id}"
                self.body = "This is a post"
                self.hidden = "This data is not exposed in the schema"
                self.keywords = ["foo", "bar", 1, True, None]

        BlogImage = GraphQLObjectType(
            "Image",
            {
                "url": GraphQLField(GraphQLString),
                "width": GraphQLField(GraphQLInt),
                "height": GraphQLField(GraphQLInt),
            },
        )

        BlogArticle: GraphQLObjectType

        BlogAuthor = GraphQLObjectType(
            "Author",
            lambda: {
                "id": GraphQLField(GraphQLString),
                "name": GraphQLField(GraphQLString),
                "pic": GraphQLField(
                    BlogImage,
                    args={
                        "width": GraphQLArgument(GraphQLInt),
                        "height": GraphQLArgument(GraphQLInt),
                    },
                    resolve=lambda obj, info, width, height: obj.pic(
                        info, width, height
                    ),
                ),
                "recentArticle": GraphQLField(BlogArticle),
            },
        )

        BlogArticle = GraphQLObjectType(
            "Article",
            {
                "id": GraphQLField(GraphQLNonNull(GraphQLString)),
                "isPublished": GraphQLField(GraphQLBoolean),
                "author": GraphQLField(BlogAuthor),
                "title": GraphQLField(GraphQLString),
                "body": GraphQLField(GraphQLString),
                "keywords": GraphQLField(GraphQLList(GraphQLString)),
            },
        )

        # noinspection PyShadowingBuiltins
        BlogQuery = GraphQLObjectType(
            "Query",
            {
                "article": GraphQLField(
                    BlogArticle,
                    args={"id": GraphQLArgument(GraphQLID)},
                    resolve=lambda _obj, _info, id: Article(id),
                ),
                "feed": GraphQLField(
                    GraphQLList(BlogArticle),
                    resolve=lambda *_args: [Article(n + 1) for n in range(10)],
                ),
            },
        )

        BlogSchema = GraphQLSchema(BlogQuery)

        # noinspection PyPep8Naming,PyMethodMayBeStatic
        class Author:
            def pic(self, info_, width: int, height: int) -> "Pic":
                return Pic(123, width, height)

            @property
            def recentArticle(self) -> Article:
                return Article(1)

        class JohnSmith(Author):
            id = 123
            name = "John Smith"

        class Pic:
            def __init__(self, uid: int, width: int, height: int):
                self.url = f"cdn://{uid}"
                self.width = f"{width}"
                self.height = f"{height}"

        document = parse(
            """
            {
              feed {
                id,
                title
              },
              article(id: "1") {
                ...articleFields,
                author {
                  id,
                  name,
                  pic(width: 640, height: 480) {
                    url,
                    width,
                    height
                  },
                  recentArticle {
                    ...articleFields,
                    keywords
                  }
                }
              }
            }

            fragment articleFields on Article {
              id,
              isPublished,
              title,
              body,
              hidden,
              notDefined
            }
            """
        )

        # Note: this is intentionally not validating to ensure appropriate
        # behavior occurs when executing an invalid query.
        assert execute_sync(schema=BlogSchema, document=document) == (
            {
                "feed": [
                    {"id": "1", "title": "My Article 1"},
                    {"id": "2", "title": "My Article 2"},
                    {"id": "3", "title": "My Article 3"},
                    {"id": "4", "title": "My Article 4"},
                    {"id": "5", "title": "My Article 5"},
                    {"id": "6", "title": "My Article 6"},
                    {"id": "7", "title": "My Article 7"},
                    {"id": "8", "title": "My Article 8"},
                    {"id": "9", "title": "My Article 9"},
                    {"id": "10", "title": "My Article 10"},
                ],
                "article": {
                    "id": "1",
                    "isPublished": True,
                    "title": "My Article 1",
                    "body": "This is a post",
                    "author": {
                        "id": "123",
                        "name": "John Smith",
                        "pic": {"url": "cdn://123", "width": 640, "height": 480},
                        "recentArticle": {
                            "id": "1",
                            "isPublished": True,
                            "title": "My Article 1",
                            "body": "This is a post",
                            "keywords": ["foo", "bar", "1", "true", None],
                        },
                    },
                },
            },
            None,
        )



import re
from typing import Optional, Tuple, cast

from .ast import Location
from .location import SourceLocation, get_location
from .source import Source


__all__ = ["print_location", "print_source_location"]


def print_location(location: Location) -> str:
    """Render a helpful description of the location in the GraphQL Source document."""
    return print_source_location(
        location.source, get_location(location.source, location.start)
    )


_re_newline = re.compile(r"\r\n|[\n\r]")


def print_source_location(source: Source, source_location: SourceLocation) -> str:
    """Render a helpful description of the location in the GraphQL Source document."""
    first_line_column_offset = source.location_offset.column - 1
    body = "".rjust(first_line_column_offset) + source.body

    line_index = source_location.line - 1
    line_offset = source.location_offset.line - 1
    line_num = source_location.line + line_offset

    column_offset = first_line_column_offset if source_location.line == 1 else 0
    column_num = source_location.column + column_offset
    location_str = f"{source.name}:{line_num}:{column_num}\n"

    lines = _re_newline.split(body)  # works a bit different from splitlines()
    location_line = lines[line_index]

    # Special case for minified documents
    if len(location_line) > 120:
        sub_line_index, sub_line_column_num = divmod(column_num, 80)
        sub_lines = [
            location_line[i : i + 80] for i in range(0, len(location_line), 80)
        ]

        return location_str + print_prefixed_lines(
            (f"{line_num} |", sub_lines[0]),
            *[("|", sub_line) for sub_line in sub_lines[1 : sub_line_index + 1]],
            ("|", "^".rjust(sub_line_column_num)),
            (
                "|",
                sub_lines[sub_line_index + 1]
                if sub_line_index < len(sub_lines) - 1
                else None,
            ),
        )

    return location_str + print_prefixed_lines(
        (f"{line_num - 1} |", lines[line_index - 1] if line_index > 0 else None),
        (f"{line_num} |", location_line),
        ("|", "^".rjust(column_num)),
        (
            f"{line_num + 1} |",
            lines[line_index + 1] if line_index < len(lines) - 1 else None,
        ),
    )


def print_prefixed_lines(*lines: Tuple[str, Optional[str]]) -> str:
    """Print lines specified like this: ("prefix", "string")"""
    existing_lines = [
        cast(Tuple[str, str], line) for line in lines if line[1] is not None
    ]
    pad_len = max(len(line[0]) for line in existing_lines)
    return "\n".join(
        prefix.rjust(pad_len) + (" " + line if line else "")
        for prefix, line in existing_lines
    )



from typing import Any

from ...error import GraphQLError
from ...language import DocumentNode, OperationDefinitionNode
from . import ASTValidationContext, ASTValidationRule

__all__ = ["LoneAnonymousOperationRule"]


class LoneAnonymousOperationRule(ASTValidationRule):
    """Lone anonymous operation

    A GraphQL document is only valid if when it contains an anonymous operation
    (the query short-hand) that it contains only that one operation definition.

    See https://spec.graphql.org/draft/#sec-Lone-Anonymous-Operation
    """

    def __init__(self, context: ASTValidationContext):
        super().__init__(context)
        self.operation_count = 0

    def enter_document(self, node: DocumentNode, *_args: Any) -> None:
        self.operation_count = sum(
            isinstance(definition, OperationDefinitionNode)
            for definition in node.definitions
        )

    def enter_operation_definition(
        self, node: OperationDefinitionNode, *_args: Any
    ) -> None:
        if not node.name and self.operation_count > 1:
            self.report_error(
                GraphQLError(
                    "This anonymous operation must be the only defined operation.", node
                )
            )
